(module wristwatch_stopwatch

   (library hiphop)

   (import
      wristwatch_beep
      wristwatch_stopwatch_datatype)

   (export stopwatch&::hiphop-program))

;%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
; The Esterel split-time stopwatch %
;%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

;; There are three submodules : a basic stopwatch, which only handles
;; the start/stop command, a lap filter, which handles the lap command,
;; and a reset handler, which determines when to reset the stopwatch.
;; These modules are put in parallel in the main STOPWATCH module,
;; with suitable renamings.


;; The basic-stopwatch& module
;;----------------------------
(define basic-stopwatch&::hiphop-program
   (hiphop&
      name: "basic-stopwatch-module"

      ;; Interface

      input:
      (event2& "HS"                   'pure)
      (event2& "start-stop-command"   'pure)

      output:
      (event2& "stopwatch-time"       'single :default (class-nil stopwatch-time))
      (event2& "stopwatch-run-status" 'single :default #f)
      (event2& "beep"                 'combined :combine combine-beeps :default no-beep-value)

      ;; Body

      (seq&
	 (emit& "stopwatch-time" (zero-stopwatch-time))

	 (loop&

	    ;; stopwatch not running
	    (emit& "stopwatch-run-status" #f)
	    (await& (now& "start-stop-command"))

	    ;; starting the stopwatch
	    (emit& "stopwatch-run-status" #t)
	    (emit& "beep" (stopwatch-beep-value))
	    (abort& (now& "start-stop-command")
	       (emit& "stopwatch-time" (increment-stopwatch-time! (preval& "stopwatch-time")))
	       (emit& "beep" (stopwatch-beep (val& "stopwatch-time")))
	       (every& (now& "HS")
		  (emit& "stopwatch-time" (increment-stopwatch-time! (preval& "stopwatch-time")))
		  (emit& "beep" (stopwatch-beep (val& "stopwatch-time")))))

	    ;; stopping the stopwatch
	    (emit& "beep" (stopwatch-beep-value))))))

;; The lap-filter& module
;;-----------------------
(define lap-filter&::hiphop-program
   (hiphop&
      name: "lap-filter-module"

      ;; Interface

      input:
      (event2& "lap-command"          'pure)
      (event2& "basic-stopwatch-time" 'single :default (class-nil stopwatch-time))

      output:
      (event2& "stopwatch-time"       'single :default (class-nil stopwatch-time))
      (event2& "stopwatch-lap-status" 'single :default #f)

      ;; Body

      (loop&
	 (emit& "stopwatch-lap-status" #f)

	 ;; not in LAP mode
	 (abort& (now& "lap-command")
	    (emit& "stopwatch-time" (val& "basic-stopwatch-time"))
	    (every& (now& "basic-stopwatch-time")
	       (emit& "stopwatch-time" (val& "basic-stopwatch-time"))))

	 ;; lap-command received, enter LAP mode
	 (emit& "stopwatch-lap-status" #t)
	 (await& (now& "lap-command")))))

;; The STOPWATCH_RESET_HANDLER module
;;-----------------------------------
(define stopwatch-reset-handler&::hiphop-program
   (hiphop&
      name: "stopwatch-reset-module"

      ;; Interface

      input:
      (event2& "start-stop-command"      'pure)
      (event2& "lap-command"             'pure)

      output:
      (event2& "reset-stopwatch-command" 'pure)

      ;; Body

      (loop&
	 (trap& reset
	    (local& ((event2& "stopwatch-stopped" 'pure))
	       (par&
		  (loop&
		     (abort& (now& "start-stop-command")
			(sustain& "stopwatch-stopped"))
		     (await& (now& "start-stop-command")))
		  (loop&
		     (await& (now& "lap-command"))
		     ;; lap-command received when not in LAP mode
		     (if& (now& "stopwatch-stopped")
			(exit& reset))
		     (await& (now& "lap-command"))))))
	 (emit& "reset-stopwatch-command"))))

;; The main STOPWATCH module
;;--------------------------
(define stopwatch&::hiphop-program
   (hiphop&
      name: "stopwatch-module"

      ;; Interface

      input:
      (event2& "HS"                   'pure)
      input:
      (event2& "start-stop-command"   'pure)
      (event2& "lap-command"          'pure)

      output:
      (event2& "stopwatch-time"       'single :default (class-nil stopwatch-time))
      output:
      (event2& "stopwatch-run-status" 'single :default #f)
      (event2& "stopwatch-lap-status" 'single :default #f)
      output:
      (event2& "beep"                 'combined :combine combine-beeps :default no-beep-value)

      ;; Body

      (local& ((event2& "reset-stopwatch-command" 'pure)
	       (event2& "basic-stopwatch-time"    'single :default (class-nil stopwatch-time)))
	 (par&
	    (loop& ; loop-each
	       (abort& (now& "reset-stopwatch-command")
		  (par&
		     (run& basic-stopwatch& bind: "stopwatch-time" "basic-stopwatch-time")
		     (run& lap-filter&))))
	    (run& stopwatch-reset-handler&)))))
