(module __hiphop_environment

   (import __hiphop_types
	   __hiphop_ast
	   __hiphop_analysis
	   __hiphop_debug
	   __hiphop_events
	   __hiphop_host_expressions
	   __hiphop_utils)

   (export (environment-initialize!::hiphop-environment% ::hiphop-program)
	   (machine-initialize! ::hiphop-machine%)
	   (environment-prepare! ::hiphop-machine%)
	   (generic clear-decorations! ::hiphop-ast%)

	   (pause-table-get proc::bstring env::hiphop-environment% id::symbol)
	   (pause-table-set! proc::bstring env::hiphop-environment% id::symbol val::bool)))

;;; Environment initialization. Takes an hiphop-program as argument
(define (environment-initialize!::hiphop-environment% program::hiphop-program)
   (with-trace 41 'environment-initialize!
      (let ((event-list '())   ; The list of all the event used in the program
	    (event-index 0)    ; A counter to uniquely index events
	    (event-table (make-hashtable))  ; The event table (index -> hiphop-event)
	    (pause-table (make-hashtable))  ; The pause table
	    (name-table  (make-hashtable))  ; The table of namespaces and programs
	    (emit-table  (make-hashtable))) ; The table of hiphop-emit objects

	 ;; Create a new-event object with the name 'event' and compute its index
	 ; :aliases is set, contains the event that the new name must aliases
	 (define (create-event event #!key aliases)
	    (cond
	       ((string? event)
		(create-event-string event :aliases aliases))
	       ((isa? event hiphop-user-event)
		(create-event-user event :aliases aliases))
	       ((isa? event hiphop-user-event2)
		(create-event-user2 event :aliases aliases))
	       (else
		(error "create-event" "illegal type" event))))

	 (define (create-event-user user-event::hiphop-user-event #!key aliases)
	    (trace-item "Creating event: " user-event
	       ", index: " event-index
	       ", aliases: " aliases)
	    (if aliases
		(let ((global-event aliases))
		   (when (string? global-event)
		      (set! global-event (assev aliases event-list)))
		   (if (and global-event (isa? global-event hiphop-event))
		       (let* ((old-event::hiphop-event global-event)
			      (new-event (instantiate::hiphop-event
					    (name user-event.name)
					    (user user-event)
					    (event% old-event.event%))))
			  new-event)
		       (begin
			  (trace-item "event-list: " event-list)
			  (trace-item "event-table: " event-table)
			  (raise (instantiate::&hiphop-compile-error
				    (proc "create-event")
				    (msg "Cannot find the aliased event")
				    (obj aliases))))))
		(let ((new-event (make-uevent& user-event event-index)))
		   (hashtable-put! event-table event-index new-event)
		   (set! event-list (cons new-event event-list))
		   (set! event-index (+fx event-index 1))
		   new-event)))

	 (define (create-event-user2 user-event::hiphop-user-event2 #!key aliases)
	    (trace-item "Creating event: " user-event
	       ", index: " event-index
	       ", aliases: " aliases)
	    (if aliases
		(let ((global-event aliases))
		   (when (string? global-event)
		      (set! global-event (assev aliases event-list)))
		   (if (and global-event (isa? global-event hiphop-event))
		       (let* ((old-event::hiphop-event global-event)
			      (new-event (instantiate::hiphop-event
					    (name user-event.name)
					    (user user-event)
					    (event% old-event.event%))))
			  new-event)
		       (begin
			  (trace-item "event-list: " event-list)
			  (trace-item "event-table: " event-table)
			  (raise (instantiate::&hiphop-compile-error
				    (proc "create-event")
				    (msg "Cannot find the aliased event")
				    (obj aliases))))))
		(let ((new-event (make-uevent2& user-event event-index)))
		   (hashtable-put! event-table event-index new-event)
		   (set! event-list (cons new-event event-list))
		   (set! event-index (+fx event-index 1))
		   new-event)))

	 (define (create-event-string event-name::bstring #!key aliases)
	    (trace-item "Creating event: " event-name
	       ", index: " event-index
	       ", aliases: " aliases)
	    (if aliases
		(let ((global-event aliases))
		   (when (string? global-event)
		      (set! global-event (assev aliases event-list)))
		   (if (and global-event (isa? global-event hiphop-event))
		       (let* ((old-event::hiphop-event global-event)
			      (new-event (instantiate::hiphop-event
					    (name event-name)
					    (user old-event.user)
					    (event% old-event.event%))))
			  new-event)
		       (begin
			  (trace-item "event-list: " event-list)
			  (trace-item "event-table: " event-table)
			  (raise (instantiate::&hiphop-compile-error
				    (proc "create-event")
				    (msg "Cannot find the aliased event")
				    (obj aliases))))))
		(let ((new-event (make-event& event-name event-index)))
		   (hashtable-put! event-table event-index new-event)
		   (set! event-list (cons new-event event-list))
		   (set! event-index (+fx event-index 1))
		   new-event)))

	 ;; A function to add pauses in the pause table
	 ; The boot pause is treated separately as it is initially selected
	 (define (register-pause! node::hiphop-ast%)
	    (with-trace 45 'register-pause!
	       (let ((id node.id))
		  (when (hashtable-contains? pause-table id)
		     (let loop ()
			(if (hashtable-contains? pause-table id)
			    (begin
			       (set! id (symbol-append id (string->symbol "-1")))
			       (loop))
			    (set! node.id id))))
		  (trace-item "Registering pause for element " id)
		  (hashtable-put! pause-table id
		     (and (isa? node hiphop-pause) (with-access::hiphop-pause node (selected?) selected?))))))

	 ;; Namespace name generator. It ensures the uniqueness of created namespaces
	 ;; Namespaces are attached to a program
	 (define (build-namespace! namespace-stack::pair program::hiphop-program)

	    ; Transforms, for instance, a list ("c" "b" "a") with joiner #\/ into "a/b/c"
	    (define (list-of-strings->string lst::pair joiner::bchar)
	       (let ((lst (reverse lst)))
		  (let loop ((l (cdr lst)) (acc (car lst)))
		     (if (pair? l)
			 (loop (cdr l) (format "~a~a~a" acc joiner (car l)))
			 acc))))

	    ; Create a free namespace name by suffixing it with an integer
	    (define (find-available-namespace)
	       (let ((tentative-namespace (list-of-strings->string namespace-stack #\/)))
		  (if (hashtable-contains? name-table tentative-namespace)
		      (let ((first (car namespace-stack))
			    (rest  (cdr namespace-stack)))
			 (let loop ((index 1))
			    (let* ((new-first (format "~a~a" first index))
				   (new-stack (cons new-first rest))
				   (new-tentative (list-of-strings->string new-stack #\/)))
			       (if (hashtable-contains? name-table new-tentative)
				   (loop (+fx index 1))
				   new-stack))))
		      namespace-stack)))

	    ; Build the namespace
	    (let* ((namespace-list (find-available-namespace))
		   (namespace-name (list-of-strings->string namespace-list #\/)))
	       (trace-item "Namespace-name: " namespace-name)
	       (hashtable-put! name-table namespace-name program)
	       namespace-list))

	 ;; Register an emit expression in the emit table
	 (define (register-emit! emit::hiphop-emit)
	    (let ((event-index emit.event.event%.index)
		  (emit-index  emit.index))
	       (let ((emit-table-key (list event-index emit-index)))
		  (if (hashtable-contains? emit-table emit-table-key)
		      (raise (instantiate::&hiphop-internal-error
				(proc "register-emit!")
				(msg  "emit expresion already exists in emit table")
				(obj  emit)))
		      (hashtable-put! emit-table emit-table-key emit)))))

	 ;; Prepare the environment
	 ; Create event objects for interface's inputs & outputs
	 (let ((default-namespace (build-namespace! (list program.name) program))
		(input-events  (map create-event program.formal-inputs))
		(output-events (map create-event program.formal-outputs))
		(extract-event ; an utility function to create an a-list of (name index)
		   (lambda (event::hiphop-event)
		      (with-access::hiphop-event event (event% name)
			 (with-access::hiphop-event% event% (index)
			    (list (string->symbol name) index))))))

	    ; Set the environment of the initial program
	    (set! program.%env event-list)

	    ; Recursively build the environment on the children, arguments are:
	    (build-environment!
	       program.reactive-code 	; The actual code
	       event-list 		; The list events (initially inputs & outputs only)
	       '() 			; The list of traps, in the form (trap-name k)
	       create-event 		; The function to create an new event (or alias)
	       register-pause! 		; The function to register a pause in the pause table
	       build-namespace! 	; The namespace builder
	       default-namespace 	; The current 'namespace', to distinguish modules
	       register-emit!) 		; The procedure to register an emit expression

	    ;; Instantiates the actual environment
	    (instantiate::hiphop-environment%
	       ; We reverse in order to preserve the user's declaration order
	       (events  (list->vector (reverse event-list))) ; WARN: Potential risk of failure here
	       ; We create a mapping from event names to indexes
	       (inputs  (map extract-event input-events))
	       ; Same for output events
	       (outputs (map extract-event output-events))
	       ; The local event table (managed by host.scm)
	       (%events (make-hashtable))
	       ; The global event table
	       (%gevents event-table)
	       ; The pause table
	       (%pauses pause-table)
	       ; The namespace/programs table
	       (%init-table name-table)
	       ; The table of emit statements
	       (%emit-table emit-table))))))

;;; Machine initialization
;; This function only calls the initialize: parameters of hiphop-programs
;; This is separated from the environment initialization, as the machine
;; is a parameter of that functions. So we need to create the machine before
;; invoking such functions
(define (machine-initialize! machine::hiphop-machine%)
   (with-trace 41 'machine-initialize!
      ;; This variable contains the events object that have to be published
      ;; on the machine input interface
      (let ((extra-events '()))

	 ;; Aliases an existing event, very similar to create-event defined above
	 (define (make-aliased-event name aliases event-list)
	    (let* ((alias-index (hashtable-get machine.environment.%events aliases))
		   (result (if (number? alias-index)
			       (hashtable-get machine.environment.%gevents alias-index)
			       (assev aliases event-list))))
	       (trace-item "result: " result "; table: " machine.environment.%events)
	       (if (and result (isa? result hiphop-event) (not (nil? result)))
		   (let* ((old-event::hiphop-event result)
			  (new-event (instantiate::hiphop-event
					(name name)
					(user (class-nil hiphop-user-event))
					(event% old-event.event%))))
		      (hashtable-put! machine.environment.%events name old-event.event%.index)
		      (set! extra-events (cons new-event extra-events)))
		   (raise (instantiate::&hiphop-compile-error
			     (proc "make-aliased-event")
			     (msg "Cannot find the aliased event")
			     (obj aliases))))))

	 ; The following function create a procedure of one argument, in charge of
	 ; returning the new event name according to the current namespace and the
	 ; local environment at initialization time.
	 ; It compute a fully qualified name compound of the current namespace and
	 ; the name of the aliased event.
	 (define (make-qualified-name namespace event-list)
	    (lambda (event)
	       (let ((fully-qualified-name (string-append namespace "#" event)))
		  (trace-item "make-qualified-name: " event " -> " fully-qualified-name)
		  (make-aliased-event fully-qualified-name event event-list)
		  fully-qualified-name)))

	 ;; For each pair of namespace/program, we invoke the initializer function
	 (hashtable-for-each
	    machine.environment.%init-table
	    (lambda (namespace::bstring value)
	       (when (and (isa? value hiphop-program) (not (nil? value)))
		  (with-access::hiphop-program value (initialized? initializer %env)
		     (when (and (not initialized?) initializer (procedure? initializer))
			(trace-item "Initializing namespace: " namespace  " " (typeof value))
			(let ((current-machine (lambda () machine))
			      (make-deep-name (make-qualified-name namespace %env)))
			   (initializer current-machine make-deep-name)
			   (set! initialized? #t)))))))

	 ;; Add the new events as inputs
	 (set! machine.inputs
	    (append! machine.inputs
	       ; The map does the same as the extract-event procedure defined
	       ; in environment-initialize!
	       (map (lambda (ev::hiphop-event)
		       (list (string->symbol ev.name) ev.event%.index))
		  extra-events))))))

;; Prepare the environment for a new reaction
(define (environment-prepare! machine::hiphop-machine%)
   (with-trace 3 'environment-prepare!
      (trace-item "current-inputs: " machine.current-inputs)
      (let ((input-event-indexes (map cadr machine.inputs)))
	 ;; Reset the status of all events according to external inputs
	 (hashtable-for-each machine.environment.%gevents
	    (lambda (index::bint event::hiphop-event)
	       (with-access::hiphop-event% event.event% (status emitter-counter index)
		  (trace-item "event #" index ": " event.name)
		  ; Reset the event%
		  (event-reset! event.event%)
		  ; Compute the new status of the event:
		  ; If the event is an input event, and it has been input;
		  ; then it is present
		  (if (member index input-event-indexes) ; it is an input event
		      (let ((input-match (assoc index machine.current-inputs)))
			 (if input-match ; it has been input
			     (let ((values (cdr input-match)))
				(if (pair? values)
				    (event-emit!/values event.event% values)
				    (event-emit! event.event%)))
			     ; Else, set the status according
			     ; to the number of emitters
			     (set! status (if (zerofx? emitter-counter) 'absent 'unknown))))
		      ; Mark 'absent every event that has no emitter,
		      ; or 'unknown otherwise
		      (set! status (if (zerofx? emitter-counter) 'absent 'unknown)))))))

      ;; Clear inputed events
      (set! machine.current-inputs '())

      ;; Clear any decorations on the tree
      (clear-decorations! machine.program.reactive-code)

      ;; Compute the list of reachable emit statements for each events
      (let ((pause-table machine.environment.%pauses))
	 (compute-selection-tree machine.program.reactive-code pause-table))

      ;; Find reachable emitters
      (mark! machine.program)
      ;; Analyse mark! results
      (hashtable-for-each machine.environment.%gevents
	 (lambda (index::bint event::hiphop-event)
	    (event-update-status! event.event%)))))


;;; Recursively builds the hiphop environment.
; The return value is a list (now empty as it is no longer used)
(define-generic (build-environment!
		   code::hiphop-ast% 		;; The reactive code
		   env::pair-nil 		;; The event list (managed as a stack)
		   trap::pair-nil 		;; The trap list (managed as a stack)
		   make-event::procedure 	;; The function to register events
		   register-pause!::procedure 	;; The function to register pause statements
		   build-namespace!::procedure 	;; Namespace builder
		   namespace::pair 		;; The current namespace (as a stack)
		   register-emit!::procedure)) 	;; The emit expression's register procedure

;;; Methods

;; Nothing... to do
(define-method (build-environment! s::hiphop-nothing env trap make-event register-pause! build-namespace! namespace register-emit!)
   '())

;; Assign an index to each emit statement, the index is relative to the emitted event
(define-method (build-environment! s::hiphop-emit env trap make-event register-pause! build-namespace! namespace register-emit!)
   (with-trace *hiphop-debug-build-env* 'build-environment!/emit
      (with-access::hiphop-emit s (name event index valued? source-location)
	 (cond
	    ; We first check that the emit is not already set up
	    ; This generally means that we do not receive a proper tree, a DAG for instance
	    ((and (isa? event hiphop-event) (or (>fx index -1) (not (nil? event))))
	     (raise (instantiate::&hiphop-internal-error
		       (proc "build-environment!/emit")
		       (msg "Already set!")
		       (obj s)
		       (location source-location))))
	    ; Events are identified by strings
	    ((string? name)
	     ; we look-up the event in the event-list, env
	     (let ((event-object (assev name env)))
		; Failing to find an event means that the event has not been declared
		(unless event-object
		   (raise (instantiate::&hiphop-compile-error
			     (proc "emit&")
			     (msg "Event is not declared")
			     (obj name)
			     (location source-location))))
		; we store the event object in the emit statement
		(set! event event-object)
		(with-access::hiphop-event event (event%)
		   (with-access::hiphop-event% event% (emitters emitter-counter
							 (event-valued? valued?))
		      ; we index of the emit statement
		      (set! index emitter-counter)
		      ; and increment the event's emit counter
		      (set! emitter-counter (+fx 1 emitter-counter))
		      ; events manage a list of its internal emitters
		      (set! emitters (cons index emitters))
		      ; valued? means that emit passes a value, e.g. (emit& "e" value)
		      (set! event-valued? valued?)
		      ; register the emit statement
		      (register-emit! s)))
		; initialize the host expression
		(when valued?
		   (hiphop-host-expression-prepare! s.expr env)))
	     '())
	    (else
	     (raise (instantiate::&hiphop-internal-error
		       (proc "Build-environment!/emit")
		       (msg "Unknown event object")
		       (obj s)
		       (location source-location))))))))

;; A pause is just registered using the adequate function
(define-method (build-environment! s::hiphop-pause env trap make-event register-pause! build-namespace! namespace register-emit!)
   (with-trace *hiphop-debug-build-env* 'build-environment!/pause
      (trace-item "pre-pause.id=" s.id)
      (register-pause! s)
      (trace-item "post-pause.id=" s.id)
      '()))

;; Nothing to do for goto-pause, as they alias a real pause statement, see above
(define-method (build-environment! s::hiphop-goto-pause env trap make-event register-pause! build-namespace! namespace register-emit!)
   '())

;; Recursively propagate into then and else branches
(define-method (build-environment! s::hiphop-if env trap make-event register-pause! build-namespace! namespace register-emit!)
   (with-access::hiphop-if s (then else)
      (hiphop-host-expression-prepare! s.test env)
      (append
	 (build-environment! then env trap make-event register-pause! build-namespace! namespace register-emit!)
	 (build-environment! else env trap make-event register-pause! build-namespace! namespace register-emit!))))

;; Suspend has an event-test and a body
(define-method (build-environment! s::hiphop-suspend env trap make-event register-pause! build-namespace! namespace register-emit!)
   (hiphop-host-expression-prepare! s.event-test env)
   (build-environment! s.body env trap make-event register-pause! build-namespace! namespace register-emit!))

;; In a sequence, we build the environment of all branches
; The code is a bit complex to aggregate all results in a list, that could be removed
(define-method (build-environment! s::hiphop-sequence env trap make-event register-pause! build-namespace! namespace register-emit!)
   (with-trace *hiphop-debug-build-env* 'build-environment!/seq
      (trace-item "seq.id=" s.id "; current-branch: " s.current-branch)
      (let ((l (vector-length s.branches))
	    (res '()))
	 (let loop ((i 0))
	    (when (<fx i l)
	       (trace-item "Building environment for branch " i)
	       (set! res
		  (append
		     res
		     (build-environment!
			(vector-ref s.branches i) env trap make-event register-pause! build-namespace! namespace register-emit!)))
	       (loop (+fx i 1))))
	 res)))

;; In a loop, we need first to build the environment for its body,
;; Then we copy the surface, and build the environment for the surface.
;; Last we make the body and the surface in sequence
(define-method (build-environment! s::hiphop-loop env trap make-event register-pause! build-namespace! namespace register-emit!)
   (let* ((body s.body)
	  (res1 (build-environment! body env trap make-event register-pause! build-namespace! namespace register-emit!)))
      (multiple-value-bind (surf pause?)
	 (hiphop-copy-surface body)
	 (let ((res2 (build-environment! surf env trap make-event register-pause! build-namespace! namespace register-emit!)))
	    (set! s.body (make-sequence& s.source-location (string->symbol (format "loop-dup-of-~s" s.id)) (list body surf)))
	    (append res1 res2)))))

;; Same as the sequence, build the environment for all branches
(define-method (build-environment! s::hiphop-parallel env trap make-event register-pause! build-namespace! namespace register-emit!)
   (let ((l (vector-length s.branches))
	 (res '()))
      (let loop ((i 0))
	 (when (<fx i l)
	    (set! res
	       (append
		  res
		  (build-environment!
		     (vector-ref s.branches i) env trap make-event register-pause! build-namespace! namespace register-emit!)))
	    (loop (+fx i 1))))
      res))
(define-method (build-environment! s::hiphop-parallel-dynamic env trap make-event register-pause! build-namespace! namespace register-emit!)
   (with-trace *hiphop-debug-build-env* 'build-environment!/dynpar
      (let ((l (vector-length s.branches))
	    (res '()))
	 (let loop ((i s.%next-initializable-index))
	    (trace-item "building env for branch " i)
	    (when (<fx i l)
	       (set! res
		  (append
		     res
		     (build-environment!
			(vector-ref s.branches i) env trap make-event register-pause! build-namespace! namespace register-emit!)))
	       (loop (+fx i 1))))
	 (set! s.%next-initializable-index l)
	 (unless s.%env-builder
	    (set! s.%env-builder
	       (lambda ()
		  (build-environment! s env trap make-event register-pause! build-namespace! namespace register-emit!))))
	 res)))

(define-method (build-environment! s::hiphop-genpar env trap make-event register-pause! build-namespace! namespace register-emit!)
   (with-trace *hiphop-debug-build-env* 'build-environment!/genpar
      (map (lambda (host-expr) (hiphop-host-expression-prepare! host-expr env)) s.host-expressions)
      (set! s.%env-builder
	 (lambda (e::hiphop-ast%)
	    (trace-item "Dumping environment")
	    (for-each (lambda (e::hiphop-event)
			 (trace-item "#" e.event%.index ": " e.name))
	       env)
	    (build-environment! e env trap make-event register-pause! build-namespace! namespace register-emit!)))
      (set! s.can-nil-instance (apply s.generator (make-list (length s.host-identifiers) #f)))
      (s.%env-builder s.can-nil-instance)))

(define-method (build-environment! s::hiphop-dyngenpar env trap make-event register-pause! build-namespace! namespace register-emit!)
   (with-trace *hiphop-debug-build-env* 'build-environment!/genpar
      ;; Build the environment of the inner genpar& form.
      (build-environment! s.genpar env trap make-event register-pause! build-namespace! namespace register-emit!)
      ;; Build the environment of the host expression
      (hiphop-host-expression-prepare! s.when env)
      ;; Build the dynamic-parallel
      (call-next-method)))

;; For trap, we add the trap name to the list of traps
;; and increment the k of all other trap's k in the list of traps
(define-method (build-environment! s::hiphop-trap env trap make-event register-pause! build-namespace! namespace register-emit!)
   (with-access::hiphop-trap s (body name)
      (let ((new-trap
	       (cons (list name 2)
		  (map (lambda (e) ; (list name index)
			  (list (car e) (+fx 1 (cadr e))))
		     trap))))
	 (build-environment! body env new-trap make-event register-pause! build-namespace! namespace register-emit!))))

;; Exit just extract the k from the list of traps and assign it in the statement
(define-method (build-environment! s::hiphop-exit env trap make-event register-pause! build-namespace! namespace register-emit!)
   (with-access::hiphop-exit s (source-location trap-name index)
      (let ((match (assoc trap-name trap)))
	 (if match
	     (begin
		(set! index (cadr match))
		'())
	     (raise (instantiate::&hiphop-compile-error
		       (proc "exit&")
		       (msg "Unknown trap name")
		       (obj trap-name)
		       (location source-location)))))))

;; Local create a new event. Then it build the environment of its body
;; with the event list extended with the newly created event
(define-method (build-environment! s::hiphop-local env trap make-event register-pause! build-namespace! namespace register-emit!)
   (with-access::hiphop-local s (event user-event body)
      (let ((local-event (make-event user-event)))
	 (set! event local-event)
	 (build-environment! body (cons local-event env) trap make-event register-pause! build-namespace! namespace register-emit!))))

;; Run is a complex form, we have to:
; - make alias for all bound events,
; - extend the namespace with the run& rename parameter, if present, or using the
;   submodule name, and guarantee the uniqueness of the extended namespace
; - initialize the submodule
(define-method (build-environment! run::hiphop-run env trap make-event register-pause! build-namespace! namespace register-emit!)
   (with-trace *hiphop-debug-build-env* 'build-environment!/run&
      (trace-fields run source-location name)
      (trace-item "program.name: " run.program.name)
      ;; First, compute the hashtable of bindings
      (for-each (lambda (couple)
		   (if (and (pair? couple) (not (pair? (cdr couple))))
		       (let ((local-name  (car couple))
			     (global-name (cdr couple)))
			  (hashtable-put! run.%bindings local-name global-name))
		       (raise (instantiate::&hiphop-internal-error
				 (proc "build-environment/run&")
				 (msg  "Malformed binding")
				 (obj  couple)))))
	 run.bindings)

      (let ((extended-env '()))
	 ;; Here the first binding resolution logic :
	 ; for each input event, do :
	 ; - check if there is a binding, and returns that binding;
	 ; - If there is an empty binding ""; make a new event and add it.
	 ; - If there is a real binding; check matching and add it.
	 ; - If there are no binding; look for a matching event in the parent env;
	 ;  - if found, do nothing; else error
	 ; same for output events
	 (define (get-binding name)
	    (hashtable-get run.%bindings name))

	 (define (make-new-event user-event)
	    (make-event user-event))

	 (define (make-aliased-event user-event alias)
	    (make-event user-event :aliases alias))

	 (define (lookup-event user-event)
	    (assev user-event env))

	 (define (resolve-binding event)
	    (cond
	       ((string? event)
		(resolve-binding-string event))
	       ((isa? event hiphop-user-event)
		(resolve-binding-user event))
	       ((isa? event hiphop-user-event2)
		(resolve-binding-user2 event))
	       (else
		(raise (instantiate::&hiphop-internal-error
			  (proc "build-environment!/run")
			  (msg "Illegal event declaration")
			  (obj event)
			  (location run.program.source-location))))))

	 (define (resolve-binding-string local-event-name::bstring)
	    (let ((binding (get-binding local-event-name)))
	       (trace-item "binding for " local-event-name ": " binding)
	       (if binding
		   (set! extended-env
		      (cons
			 (if (string=? binding "")
			     (make-new-event local-event-name)
			     (let ((global-event (lookup-event binding)))
				(if global-event
				    (make-aliased-event local-event-name global-event)
				    (raise (instantiate::&hiphop-compile-error
					      (proc "run&")
					      (msg  "Illegal binding. Unknown event")
					      (obj  local-event-name)
					      (location run.source-location))))))
			 extended-env))
		       (let ((global-event (lookup-event local-event-name)))
			  (unless global-event
			     (raise (instantiate::&hiphop-compile-error
				       (proc "run&")
				       (msg  "Illegal binding. Unknown event")
				       (obj  local-event-name)
				       (location run.source-location))))))))

	 (define (resolve-binding-user local-user-event::hiphop-user-event)
	    (let ((binding (get-binding local-user-event.name)))
	       (trace-item "binding for " local-user-event ": " binding)
	       (if binding
		   (set! extended-env
		      (cons
			 (if (string=? binding "")
			     (make-new-event local-user-event)
			     (let ((global-event (lookup-event binding)))
				(if global-event
				    (begin
				       (check-event-match local-user-event global-event)
				       (make-aliased-event local-user-event global-event))
				    (raise (instantiate::&hiphop-compile-error
					      (proc "run&")
					      (msg  "Illegal binding. Unknown event")
					      (obj  local-user-event)
					      (location run.source-location))))))
			 extended-env))
		   (let ((global-event (lookup-event local-user-event.name)))
		      (if global-event
			  (check-event-match local-user-event global-event)
			  (raise (instantiate::&hiphop-compile-error
				    (proc "run&")
				    (msg  "Illegal binding. Unknown event")
				    (obj  local-user-event)
				    (location run.source-location))))))))

	 (define (resolve-binding-user2 local-user-event::hiphop-user-event2)
	    (let ((binding (get-binding local-user-event.name)))
	       (trace-item "binding for " local-user-event ": " binding)
	       (if binding
		   (set! extended-env
		      (cons
			 (if (string=? binding "")
			     (make-new-event local-user-event)
			     (let ((global-event (lookup-event binding)))
				(if global-event
				    (begin
				       (check-event-match2 local-user-event global-event)
				       (make-aliased-event local-user-event global-event))
				    (raise (instantiate::&hiphop-compile-error
					      (proc "run&")
					      (msg  "Illegal binding. Unknown event")
					      (obj  local-user-event)
					      (location run.source-location))))))
			 extended-env))
		   (let ((global-event (lookup-event local-user-event.name)))
		      (if global-event
			  (check-event-match2 local-user-event global-event)
			  (raise (instantiate::&hiphop-compile-error
				    (proc "run&")
				    (msg  "Illegal binding. Unknown event")
				    (obj  local-user-event)
				    (location run.source-location))))))))

	 (define (check-event-match uevent::hiphop-user-event other::hiphop-event)
	    (when (isa? other.user hiphop-user-event2)
	       (raise (hiphop-compile-error run.source-location
			 "run&"
			 "Attempting to bind a event& with a event2& object. Aborting"
			 (list uevent other.user))))
	    (let ((o-uevent::hiphop-user-event other.user))
	       (unless (eq? uevent.combine o-uevent.combine)
		  (raise (instantiate::&hiphop-compile-error
			    (proc "run&")
			    (msg "Illegal binding. Combine function of bound events do not match.")
			    (obj (list uevent o-uevent))
			    (location run.source-location))))
	       (unless (eq? uevent.default o-uevent.default)
		  (raise (instantiate::&hiphop-compile-error
			    (proc "run&")
			    (msg "Illegal binding. Default value of bound events do not match.")
			    (obj (list uevent o-uevent))
			    (location run.source-location))))
	       (unless (eq? uevent.multi? o-uevent.multi?)
		  (raise (instantiate::&hiphop-compile-error
			    (proc "run&")
			    (msg "Illegal binding. One event is multi-valued, the other is not.")
			    (obj (list uevent o-uevent))
			    (location run.source-location))))))

	 (define (check-event-match2 uevent::hiphop-user-event2 other::hiphop-event)
	    (unless (uevent-equal? uevent other.user)
	       (when uevent.source-location
		  (hiphop-warn uevent.source-location "run&"
		     "Event '" uevent.name "' was declared here."))
	       (when (isa? other.user hiphop-user-event2)
		  (let ((ouevent::hiphop-user-event2 other.user))
		     (when ouevent.source-location
			(hiphop-warn ouevent.source-location "run&"
			   "Event '" ouevent.name "' was declared here."))))
	       (raise (hiphop-compile-error
			 run.source-location
			 "run&"
			 "Illegal binding. Event declarations do not match"
			 (list uevent.name other.name)))))

	 (trace-item "Extended events: " extended-env)

	 (for-each resolve-binding run.program.formal-inputs)
	 (for-each resolve-binding run.program.formal-outputs)

	 (trace-item "Extended events: " extended-env)

	 ;; Build the new environment and new namespace
	 (let ((new-namespace (build-namespace!
				 (cons (or run.name run.program.name) namespace)
				 run.program))
	       (new-env (append extended-env env)))

	    (set! run.program.%env new-env)

	    ;; Recursively build the environment of the submodule with the extended parameters
	    ; Note: The trap list is reset to '() here, as you should not exit a trap
	    ; which is not defined in the submodule
	    (build-environment! run.program.reactive-code
	       new-env '() make-event register-pause! build-namespace! namespace register-emit!)))))

;; Nothing to do for atoms
(define-method (build-environment! s::hiphop-atom env trap make-event register-pause! build-namespace! namespace register-emit!)
   (hiphop-host-expression-prepare! s.expr env)
   '())

;;; Clear decorations: remove any decorations on the tree by depth-first traversal.
;; Selection will be computed just after by compute-selection-tree defined below.
(define-generic (clear-decorations! ast::hiphop-ast%))

(define-method (clear-decorations! nop::hiphop-nothing)
   #unspecified)
(define-method (clear-decorations! emit::hiphop-emit)
   (set! emit.reachable? #f)
   (set! emit.flagged? #f)
   (when emit.valued?
      (hiphop-host-expression-reset! emit.expr)))
(define-method (clear-decorations! pause::hiphop-pause)
   ; Unselect the pause
   (set! pause.selected? #f))
(define-method (clear-decorations! pause::hiphop-goto-pause)
   #unspecified)
(define-method (clear-decorations! if%::hiphop-if)
   ; We forget that the test has been computed
   (set! if%.branch 'none)
   (hiphop-host-expression-reset! if%.test)
   (clear-decorations! if%.then)
   (clear-decorations! if%.else))
(define-method (clear-decorations! susp::hiphop-suspend)
   (hiphop-host-expression-reset! susp.event-test)
   (clear-decorations! susp.body))
(define-method (clear-decorations! seq::hiphop-sequence)
   ; The sequence restarts from the first branch
   (set! seq.current-branch -1)
   (my-vector-for-each clear-decorations! seq.branches))
(define-method (clear-decorations! loop::hiphop-loop)
   (clear-decorations! loop.body))
(define-method (clear-decorations! par::hiphop-parallel)
   ; Mark all branches as active
   (set! par.active-branches (make-vector (vector-length par.branches) #t))
   (my-vector-for-each clear-decorations! par.branches))
(define-method (clear-decorations! genpar::hiphop-genpar)
   (map hiphop-host-expression-reset! genpar.host-expressions)
   (clear-decorations! genpar.can-nil-instance)
   (call-next-method))
(define-method (clear-decorations! dgp::hiphop-dyngenpar)
   (set! dgp.expanded? #f)
   (set! dgp.selected-branches? #f)
   (set! dgp.has-new-branch? #f)
   (hiphop-host-expression-reset! dgp.when)
   (call-next-method))
(define-method (clear-decorations! trap::hiphop-trap)
   (clear-decorations! trap.body))
(define-method (clear-decorations! exit::hiphop-exit)
   #unspecified)
(define-method (clear-decorations! local::hiphop-local)
   (clear-decorations! local.body))
(define-method (clear-decorations! run::hiphop-run)
   (clear-decorations! run.program.reactive-code))
(define-method (clear-decorations! atom::hiphop-atom)
   (hiphop-host-expression-reset! atom.expr)
   ; We forget that the atom has been ran
   (set! atom.ran? #f))

;;; Compute the selection tree, using depth-first traversal
; Returns a boolean saying if the current path is selected or not
; - instantaneous statements returns #f
; - selected `pause's return #t
; - unselected `pause's return #f
; - if, sequence, parallels select a path according to its selected children
(define-generic (compute-selection-tree::bool ast::hiphop-ast% pause-table))

(define-method (compute-selection-tree nop::hiphop-nothing pause-table) #f)
(define-method (compute-selection-tree emit::hiphop-emit pause-table)
   (with-trace *hiphop-debug-compute-sel-tree* 'compute-selection-tree/emit&
      (trace-fields emit id)
      #f))
(define-method (compute-selection-tree pause::hiphop-pause pause-table)
   (with-trace *hiphop-debug-compute-sel-tree* 'compute-selection-tree/pause&
      ;; Sanity check: ensure that the pause id is in the table
      (unless (hashtable-contains? pause-table pause.id)
	 (raise (instantiate::&hiphop-internal-error
		   (proc "compute-selection-tree")
		   (msg "Unknown pause id in pause table")
		   (obj pause))))
      ;; Set pause.selected? value according to the table
      (set! pause.selected? (hashtable-get pause-table pause.id))
      (trace-fields pause id selected? source-location)
      ; Pause can be crossed iff it has been previously activated
      pause.selected?))
(define-method (compute-selection-tree pause::hiphop-goto-pause pause-table)
   (with-trace *hiphop-debug-compute-sel-tree* 'compute-selection-tree/goto-pause&
      ;; Sanity check: ensure that the pause id is in the table
      (unless (hashtable-contains? pause-table pause.id)
	 (raise (instantiate::&hiphop-internal-error
		   (proc "compute-selection-tree")
		   (msg "Unknown pause id in pause table")
		   (obj (list pause pause-table)))))
      (trace-item "pause.id: " pause.id ", selected? " (hashtable-get pause-table pause.id))
      ;; We never restart from a goto-pause
      #f))
(define-method (compute-selection-tree if%::hiphop-if pause-table)
   (with-trace *hiphop-debug-compute-sel-tree* 'compute-selection-tree/if&
      (trace-fields if% id)
      (let ((r-then (compute-selection-tree if%.then pause-table))
	    (r-else (compute-selection-tree if%.else pause-table)))
	 (when (or r-then r-else)
	    (cond
	       (r-then (set! if%.branch 'then))
	       (r-else (set! if%.branch 'else))
	       (else
		(raise (instantiate::&hiphop-internal-error
			  (proc "compute-selection-tree")
			  (msg "If statement with selected then and else branches")
			  (obj if%)))))
	    (trace-fields if% branch)
	    #t))))
(define-method (compute-selection-tree susp::hiphop-suspend pause-table)
   (with-trace *hiphop-debug-compute-sel-tree* 'compute-selection-tree/suspend&
      (trace-fields susp id)
      (let ((body-selected? (compute-selection-tree susp.body pause-table)))
	 (trace-item "susp.body-selected? " body-selected?)
	 body-selected?)))
(define-method (compute-selection-tree seq::hiphop-sequence pause-table)
   (with-trace *hiphop-debug-compute-sel-tree* 'compute-selection-tree/seq&
      (trace-fields seq id)
      (let ((len (vector-length seq.branches)))
      	 (let loop ((current 0))
      	    (when (<fx current len)
      	       (let ((branch (vector-ref seq.branches current)))
      		  (let ((res (compute-selection-tree branch pause-table)))
		     (trace-item "Checking branch " current "; res=" res)
      		     (if res
			 (if (>fx seq.current-branch -1)
			     (raise (instantiate::&hiphop-internal-error
					  (proc "compute-selection-tree/seq&")
					  (msg "More than one active branch in sequence")
					  (obj seq)))
			     (begin
				(trace-item "Setting current branch to: " current)
				(set! seq.current-branch current)
				(loop (+fx current 1))))
      			 (loop (+fx current 1))))))))
      (trace-fields seq current-branch)
      ; Return #t if the statement has one active branch
      (>fx seq.current-branch -1)))
(define-method (compute-selection-tree loop::hiphop-loop pause-table)
   (with-trace *hiphop-debug-compute-sel-tree* 'compute-selection-tree/loop&
      (trace-fields loop id)
      (compute-selection-tree loop.body pause-table)))
(define-method (compute-selection-tree par::hiphop-parallel pause-table)
   (with-trace *hiphop-debug-compute-sel-tree* 'compute-selection-tree/par&
      (trace-fields par id active-branches)
      (let ((res (map (lambda (branch)
			 (compute-selection-tree branch pause-table))
		    (vector->list par.branches))))
	 (trace-item "active-branches: " res)
	 ; If some branch are actives, returns #t
	 (when (any? (lambda (x) x) res)
	    (set! par.active-branches (list->vector res))
	    #t))))
(define-method (compute-selection-tree genpar::hiphop-genpar pause-table)
   (with-trace *hiphop-debug-compute-sel-tree* 'compute-selection-tree/genpar&
      (trace-fields genpar id)
      (if genpar.expanded?
	  (call-next-method)
	  (compute-selection-tree genpar.can-nil-instance pause-table))))
(define-method (compute-selection-tree dgp::hiphop-dyngenpar pause-table)
   (with-trace *hiphop-debug-compute-sel-tree* 'compute-selection-tree/dyngenpar&
      (trace-fields dgp id has-started?)
      ;; Compute selection tree of current branches
      (set! dgp.selected-branches? (call-next-method))
      (trace-item "selected-branches? " dgp.selected-branches?)
      ;; We always returns true as soon as a dyngenpar has received Go
      ;; once, as it always await for its when condition
      (or dgp.selected-branches? dgp.has-started?)))
(define-method (compute-selection-tree trap::hiphop-trap pause-table)
   (with-trace *hiphop-debug-compute-sel-tree* 'compute-selection-tree/trap&
      (trace-fields trap id name)
      (compute-selection-tree trap.body pause-table)))
(define-method (compute-selection-tree exit::hiphop-exit pause-table)
   (with-trace *hiphop-debug-compute-sel-tree* 'compute-selection-tree/exit&
      (trace-fields exit id trap-name)
      #f))
(define-method (compute-selection-tree local::hiphop-local pause-table)
   ;(with-trace *hiphop-debug-compute-sel-tree* 'compute-selection-tree/local&
   ;   (trace-fields local id)
      (compute-selection-tree local.body pause-table));)
(define-method (compute-selection-tree run::hiphop-run pause-table)
   (with-trace *hiphop-debug-compute-sel-tree* 'compute-selection-tree/run&
      (trace-fields run id name)
      (when run.program.name
	 (trace-item "run.program.name: " run.program.name))
      (let ((program-selected?
	       (compute-selection-tree run.program.reactive-code pause-table)))
	 (trace-item "run.program.selected? " program-selected?)
	 program-selected?)))
(define-method (compute-selection-tree atom::hiphop-atom pause-table)
   (with-trace *hiphop-debug-compute-sel-tree* 'compute-selection-tree/atom&
      (trace-fields atom id)
      #f))

;;; Pause table management
;; Setter for the pause table, with all needed sanity checks
(define (pause-table-set! proc::bstring env::hiphop-environment% pause-id::symbol val::bool)
   (when (nil? env)
      (raise (instantiate::&hiphop-internal-error
		(proc proc)
		(msg "Called with a nil environment")
		(obj env))))
   (let ((pause-table env.%pauses))
      (unless (hashtable? pause-table)
	 (raise (instantiate::&hiphop-internal-error
		   (proc proc)
		   (msg "Top-level pause-table is not an hashtable")
		   (obj pause-table))))
      (unless (hashtable-contains? pause-table pause-id)
	 (raise (instantiate::&hiphop-internal-error
		   (proc proc)
		   (msg "Unknown pause id in pause table")
		   (obj pause-id))))
      (hashtable-put! pause-table pause-id val)))
;; Getter for the pause table, with all needed sanity checks
(define (pause-table-get proc::bstring env::hiphop-environment% pause-id::symbol)
   (when (nil? env)
      (raise (instantiate::&hiphop-internal-error
		(proc proc)
		(msg "Called with a nil environment")
		(obj env))))
   (let ((pause-table env.%pauses))
      (unless (hashtable? pause-table)
	 (raise (instantiate::&hiphop-internal-error
		   (proc proc)
		   (msg "Top-level pause-table is not an hashtable")
		   (obj pause-table))))
      (unless (hashtable-contains? pause-table pause-id)
	  (raise (instantiate::&hiphop-internal-error
		       (proc proc)
		       (msg "Unknown pause id in pause table")
		       (obj pause-id))))
      (hashtable-get pause-table pause-id)))
