(module __hiphop_host_expressions

   (import
      __hiphop_types
      __hiphop_events
      __hiphop_debug
      __hiphop_utils)

   (export
      (hiphop-host-expression-prepare! expr::hiphop-host-expr env::pair-nil)
      (hiphop-host-expression-resolve!::bool expr::hiphop-host-expr env::hiphop-environment%)
      (hiphop-host-expression-eval! expr::hiphop-host-expr env::hiphop-environment%)
      (hiphop-host-expression-compute! expr::hiphop-host-expr env::hiphop-environment%)
      (hiphop-host-expression-computed?::bool expr::hiphop-host-expr)
      (hiphop-host-expression-reset! expr::hiphop-host-expr)))

(define (make-event-resolver proc-name::bstring expr::hiphop-host-expr env::pair-nil)
   (lambda (event/location)
      (let ((event-name
	       (if (pair? event/location) ;; FIXME Backward compat with let&
		   (car event/location)
		   event/location)))
	 (let ((event-object (assev event-name env)))
	    (if event-object
		(let ((event::hiphop-event event-object))
		   (hashtable-put! expr.%event-table event-name event.event%.index)
		   event.event%.index)
		(raise (instantiate::&hiphop-compile-error
			  (proc proc-name)
			  (msg "Event is not declared")
			  (obj event-name)
			  (location (if (pair? event/location)
					(cdr event/location)
					expr.source-location)))))))))

;; Exported
(define (hiphop-host-expression-prepare! expr::hiphop-host-expr env::pair-nil)
   (with-trace *hiphop-debug-host-expr* 'hiphop-host-expression-prepare!
      (trace-fields expr source-location)
      ;; Compute synchronization vector for now& events and add then to the event table
      (set! expr.%now-vindexes
	 (list->vector (map (make-event-resolver "now&" expr env) expr.%now-depends)))
      ;; Compute synchronization vector for val& events and add them to the event table
      (set! expr.%val-vindexes
	 (list->vector (map (make-event-resolver "val&" expr env) expr.%val-depends)))
      ;; Add pre/preval events to the event table
      (for-each (make-event-resolver "pre&/preval&" expr env) expr.%pre-depends)
      ;; Nothing left to do
      #unspecified))

;;; Utilities

;;; Exported
;; Given an expression and an environment, this procedure checks that
;; all dependencies are satisfied for the host expression
(define (hiphop-host-expression-resolve!::bool expr::hiphop-host-expr env::hiphop-environment%)
   (with-trace *hiphop-debug-host-expr* 'hiphop-host-expression-resolve!
      (trace-fields expr source-location %now-vindexes %val-vindexes %pre-depends)

      (unless expr.%resolved?
	 (let ((now-unresolved? #f)  ; unresolved events in the sense of now&
	       (val-unresolved? #f)) ; unresolved events in the sense of val&&
	    ;; Check if some events are not resolved for now&
	    (my-vector-for-each
	       (lambda (index)
		  (when (unresolved?/now index env)
		     (set! now-unresolved? #t)))
	       expr.%now-vindexes)
	    ;; True means that some event presences are not determined yet
	    (unless now-unresolved?
	       ;; Continue with event values
	       (my-vector-for-each
		  (lambda (index)
		     (when (unresolved?/val index env)
			(set! val-unresolved? #t)))
		  expr.%val-vindexes)
	       ;; True means that some event values are not determined yet
	       (unless val-unresolved?
		  ;; Mark the expression as resolved, but does not compute the value yet
		  (set! expr.%resolved? #t)))))
      (trace-item "resolved? " expr.%resolved?)
      expr.%resolved?))

;; Exported
(define (hiphop-host-expression-eval! expr::hiphop-host-expr env::hiphop-environment%)
   (with-trace *hiphop-debug-host-expr* 'hiphop-host-expression-eval!
      (unless expr.%resolved?
	 (raise (instantiate::&hiphop-internal-error
		   (proc "hiphop-host-expression-eval!")
		   (msg "Attempting to evaluate a non resolved expression")
		   (obj expr))))
      (unless expr.%computed?
	 ;; We need to setup the environment in the machine
	 (let ((old-event-table env.%events))
	    (set! env.%events expr.%event-table)
	    (unwind-protect
	       (let ((result (expr.expression env.current-machine)))
		  (when (eq? result '())
		     (trace-item expr))
		  (set! expr.%computed? #t)
		  (set! expr.%computed-expression result))
	       (set! env.%events old-event-table))))
      (trace-item "value? " expr.%computed-expression)
      expr.%computed-expression))

;; A predicate to check is an event is not resolved in the sense of now&
(define (unresolved?/now::bool event-index::bint env::hiphop-environment%)
   (unless (hashtable-contains? env.%gevents event-index)
      (raise (instantiate::&hiphop-internal-error
		(proc "hiphop-host-expression-resolve!")
		(msg "Unknown event index")
		(obj event-index))))
   (let ((event-object::hiphop-event (hashtable-get env.%gevents event-index)))
      (event-update-status! event-object.event%)
      ;; An event presence is determined if either:
      ;; - it is already present or absent (external input event)
      ;; - it is not reachable
      ;; - it is reachable and one emitter has been emitted
      (when (and (eq? event-object.event%.status 'unknown)
		 (>fx event-object.event%.reachable-count 0)
		 (zerofx? event-object.event%.emitted-count))
	 (trace-item "Undetermined presence for event: " event-object)
	 #t)))

;; A predicate to check is an event un not resolved in the sens of val&
(define (unresolved?/val::bool event-index::bint env::hiphop-environment%)
   (unless (hashtable-contains? env.%gevents event-index)
      (raise (instantiate::&hiphop-internal-error
		(proc "hiphop-host-expression-resolve!")
		(msg "Unknown event index")
		(obj event-index))))
   (let ((event-object::hiphop-event (hashtable-get env.%gevents event-index)))
      (event-update-status! event-object.event%)
      ;; An event value is determined if the reachable count equals the emitted count
      (unless (=fx event-object.event%.reachable-count event-object.event%.emitted-count)
	 (trace-item "Undetermined value for event: " event-object)
	 #t)))

;; Combine the actions of -resolve! and -eval!, returns (values resolved? value)
(define (hiphop-host-expression-compute! expr::hiphop-host-expr env::hiphop-environment%)
   (if (hiphop-host-expression-resolve! expr env)
       (values #t (hiphop-host-expression-eval! expr env))
       (values #f #f)))

(define (hiphop-host-expression-computed? expr::hiphop-host-expr)
   expr.%computed?)

(define (hiphop-host-expression-reset! expr::hiphop-host-expr)
   (set! expr.%resolved? #f)
   (set! expr.%computed? #f))
