(module __hiphop_machine

   (import  __hiphop_ast
	    __hiphop_debug
	    __hiphop_dom
	    __hiphop_environment
	    __hiphop_events
	    __hiphop_must&can
	    __hiphop_types
	    __hiphop_utils)

   (export  (make-hiphop-machine::hiphop-machine% ::hiphop-program #!key auto-react? raw-errors? dynamic?)
	    (hiphop-react! ::hiphop-machine%)
	    (hiphop-input! ::hiphop-machine% ::bstring . values)
	    (hiphop-input-and-react! e::hiphop-machine% ::bstring . values)
	    (hiphop-add-event-listener! ::hiphop-machine% event ::procedure)
	    (hiphop-append! ::hiphop-machine% ::hiphop-program)
	    (hiphop-now& ::hiphop-machine% ::bstring)
	    (hiphop-val& ::hiphop-machine% ::bstring)
	    (hiphop-pre& ::hiphop-machine% ::bstring)
	    (hiphop-preval& ::hiphop-machine% ::bstring)))

(define (convert-error e::&hiphop-error)
   (cond-expand
      (scheme2js
	 (let ((err (js-new (@ Error js) e.msg)))
	    (set! err.hopLocation (symbol->string e.constructor.sc_name))
	    (set! err.name e.proc)
	    (set! err.scObject e.obj)
	    (when (pair? e.location)
	       (let ((file-char (cdr e.location)))
		  (set! err.fileName (car file-char))
		  (set! err.charNumber (cadr file-char))))
	    (set! err.cause e)
	    err))
      (else
       e)))

;;; Create an hiphop reactive machine with the code of program
;; auto-react?: if true, it means that if hiphop-input! is invoked during a reaction,
;; it will re-call itself at the end of that reaction. Default: false.
(define (make-hiphop-machine::hiphop-machine% program::hiphop-program #!key auto-react? raw-errors? dynamic?)
   (with-trace *hiphop-debug-make-machine* 'make-hiphop-machine
      (with-handler
	 (lambda (e)
	    (if (and (isa? e &hiphop-error) (not raw-errors?))
		(raise (convert-error e))
		(raise e)))

	 ;; First, create a copy of the program
	 (let ((program-copy::hiphop-program (hiphop-copy program)))
	    (with-access::hiphop-program program-copy (reactive-code source-location name
							 formal-inputs formal-outputs)
	       ; Then, nest the program's copy inside (seq& boot-pause <copy> (halt&))
	       ; TODO: Move this in a dedicated function in ast.scm? named hiphop-finalize?
	       (let ((seq-halted-code
			(instantiate::hiphop-sequence
			   (source-location source-location)
			   (id 'seq/halt)
			   (branches
			      (vector
				 ;; Boot pause
				 (instantiate::hiphop-pause
				    (selected? #t) ; already selected
				    (id '__boot)
				    (source-location source-location))
				 ;; Reactive code
				 (if dynamic?
				     (instantiate::hiphop-parallel-dynamic
					(source-location source-location)
					(id '__dynpar)
					(active-branches '#(#f))
					(branches
					   (vector
					      reactive-code)))
				     reactive-code)
				 ;; Halt loop
				 (instantiate::hiphop-loop
				    (source-location source-location)
				    (id '__halt_loop)
				    (body (instantiate::hiphop-pause
					     (source-location source-location)
					     (id '__halt_pause)))))))))
		  (set! reactive-code seq-halted-code))

	       (let* (; Now, create the environment
		      (env::hiphop-environment% (environment-initialize! program-copy))
		      ; and the reactive machine
		      (new-machine (instantiate::hiphop-machine%
				      (program program-copy)
				      (inputs env.inputs)
				      (outputs env.outputs)
				      (current-inputs '())
				      (state 'ready)
				      (environment env)
				      (auto-react? auto-react?)
				      (raw-errors? raw-errors?)
				      (dynamic? dynamic?))))
		  ; The following code allow, for _client code only_,  the use of
		  ; (add-event-listener! <machine> <event> <callback> <capture>)
		  ; The capture parameter is ignored and not used at all
		  (cond-expand
		     (scheme2js
			(js-field-set! new-machine "hop_add_event_listener"
			   (lambda (event proc capture)
			      (hiphop-add-event-listener! new-machine event
				 (lambda (value)
				    (let ((evt (js-new (@ HopEvent js) event value)))
				       (proc evt))))))))

		  ; Initialize the machine (invokes program initializers)
		  (machine-initialize! new-machine)
		  ; Register the machine in the environment
		  (set! env.current-machine new-machine)
		  ; Return that machine
		  new-machine))))))

;;; Compute a reaction (or many, depending on what may happen during one reaction)
(define (hiphop-react! machine::hiphop-machine%)
   (with-trace *hiphop-debug-react* 'hiphop-react!
      (with-handler
	 (lambda (e)
	    (if (and (isa? e &hiphop-error) (not machine.raw-errors?))
		(raise (convert-error e))
		(raise e)))

	 (case machine.state
	    ; If the machine is already running, register the hiphop-react! request
	    ((running)
	     (trace-item "Already reacting. Registering request.")
	     (set! machine.must-react? #t))

	    ; The reaction starts here
	    ((ready)
	     (set! machine.must-react? #f)
	     (set! machine.state 'running)

	     ;; Initialize the environment for the upcoming reaction
	     (environment-prepare! machine)

	     ;; Compute the reaction
	     (React! machine.program machine.environment)

	     ;; Reaction is finished, invoke the listeners
	     ; The invoke-listener function (defined below) will actually check
	     ; that an output event is present.
	     (for-each invoke-listener
		(map (lambda (pair-name-index) (hashtable-get machine.environment.%gevents (cadr pair-name-index)))
		   machine.environment.outputs))

	     ;; Check if there is programs to be appended
	     (unless (null? machine.dynprog)
		(for-each (lambda (program) (%machine-append-program! machine program)) machine.dynprog)
		(set! machine.dynprog '()))

	     ;; End of the reaction
	     (set! machine.state 'ready)

	     ;; Start a new reaction if we received hiphop-react! while already running
	     ;; or the machine is in auto-react? mode and we got hiphop-input!
	     (when machine.must-react?
		(trace-item "must-react? flag is set to true; restarting.")
		(hiphop-react! machine)))))))

;;; Register an external input named 'event' with values 'values'
; e.g. (hiphop-input! <machine> "foo" 1 2 3) mark the "foo" event present
; for the next reaction; "foo" is multi-valued and its values are the one
; of the multi-set {{1, 2, 3}}
(define (hiphop-input! machine::hiphop-machine% event . values)
   (with-trace *hiphop-debug-interface* 'hiphop-input!
      (with-handler
	 (lambda (e)
	    (if (and (isa? e &hiphop-error) (not machine.raw-errors?))
		(raise (convert-error e))
		(raise e)))

	 (trace-item "Input: " event "; value: " values)
	 ; Event names are internally managed as symbol
	 (let ((event-name (string->symbol event)))
	    (with-access::hiphop-machine% machine (current-inputs)
	       ; We first resolve the event index and check its validity
	       (trace-item "inputs: " machine.inputs)
	       (let* ((event-id (assoc event-name machine.inputs))
		      (index (and event-id (cadr event-id))))
		  (when (and (number? index) (pair? values))
		     (let ((hevent (hashtable-get machine.environment.%gevents index)))
			(when (isa? hevent hiphop-event)
			   (with-access::hiphop-event hevent (event% user)
			      (when (isa? event% hiphop-pure-event)
				 (when (isa? user hiphop-user-event2)
				    (let ((uevent::hiphop-user-event2 user))
				       (when uevent.source-location
					  (hiphop-warn uevent.source-location
					     "hiphop-input!"
					     "Event declaration is here -- "
					     event))))
				 (raise (instantiate::&hiphop-interface-error
					   (proc "hiphop-input!")
					   (msg "Event has been declared pure, cannot input")
					   (obj event))))))))
		  (if index
		      ; We check if the event as already been input;
		      (let ((already-input (assoc index current-inputs)))
			 (if already-input
			     ; if true, we append the new values to the already input values
			     (append! already-input values)
			     ; else, we merely add that event in the current input event list
			     (set! current-inputs
				(cons (cons index values) current-inputs))))
		      (raise (instantiate::&hiphop-interface-error
				(proc "hiphop-input!")
				(msg "Event is not an input event")
				(obj event)))))))
	 ; If the machine is set in auto-react mode, and if the machine is currently running
	 ; We tell the machine to restart at the end of the current reaction
	 (when (and machine.auto-react? (eq? machine.state 'running))
	    (trace-item "Received input while reacting with auto-react mode. "
	       "Registering request.")
	    (set! machine.must-react? #t)))))

;;; An practical function that combine the behavior of input! and react!
(define (hiphop-input-and-react! machine::hiphop-machine% event . values)
   (apply hiphop-input! machine event values)
   (hiphop-react! machine))

;;; A procedure to add event listeners to output events
;; Multiple listeners can be added for the same event. They will be invoked in LIFO order.
;; A listener must be a procedure of one argument
;; There is currently no remove-event-listener procedure
(define (hiphop-add-event-listener! machine::hiphop-machine% event proc::procedure)
   (let ((event-name (string->symbol event)))
      (let* ((event-id (assoc event-name machine.outputs))
	     (index (and event-id (cadr event-id))))
	 (if (and (number? index) (hashtable-contains? machine.environment.%gevents index))
	     (let ((event::hiphop-event (hashtable-get machine.environment.%gevents index)))
		(set! event.event%.listeners (cons proc event.event%.listeners)))
	     (raise (instantiate::&hiphop-interface-error
		       (proc "hiphop-add-event-listener!")
		       (msg "Event in not an ouput event")
		       (obj event)))))))

;;; A procedure to add new programs into a machine
(define (hiphop-append! machine::hiphop-machine% program::hiphop-program)
   (with-trace *hiphop-debug-interface* 'hiphop-append!
      (unless machine.dynamic?
	 (raise (instantiate::&hiphop-interface-error
		   (proc "hiphop-append!")
		   (msg "Machine has not been created using :dynamic? #t")
		   (obj #f))))
      (trace-item "Appending program: " program)
      (if (eq? machine.state 'ready)
	  (%machine-append-program! machine program)
	  (set! machine.dynprog (cons program machine.dynprog)))
      #unspecified))

;;; Fetches the presence of the given event.
; This procedure is not intended to be called directly. now& is expended into this form.
(define (hiphop-now& machine::hiphop-machine% event-name::bstring)
   (let ((event-object::hiphop-event (get-event-by-name machine event-name "now&")))
      (event-present? event-object.event%)))

;;; Fetches the value of the given event.
; This procedure is not intended to be called directly. val& is expended into this form.
(define (hiphop-val& machine::hiphop-machine% event-name::bstring)
   (let ((event-object::hiphop-event (get-event-by-name machine event-name "val&")))
      (event-value event-object.event%)))

;;; Fetches the previous presence of the given event.
; This procedure is not intended to be called directly. pre& is expended into this form.
(define (hiphop-pre& machine::hiphop-machine% event-name::bstring)
   (let ((event-object::hiphop-event (get-event-by-name machine event-name "pre&")))
      (event-previously-present? event-object.event%)))

;;; Fetches the previous value of the given event.
; This procedure is not intended to be called directly. preval& is expended into this form.
(define (hiphop-preval& machine::hiphop-machine% event-name::bstring)
   (let ((event-object::hiphop-event (get-event-by-name machine event-name "preval&")))
      (event-previous-value event-object.event%)))

;;; Invoke listeners
(define (invoke-listener event-object::hiphop-event)
   ; Sanity check
   (unless (isa? event-object hiphop-event)
      (raise (instantiate::&hiphop-internal-error
		(proc "invoke-listener")
		(msg "Not an hiphop-event object")
		(obj event-object))))
   (with-trace 3 'invoke-listener
      (let ((event%::hiphop-event% event-object.event%))
	 (when (and (event-present? event%)
		    (pair? event%.listeners))
	    (for-each
	       (lambda (proc)
		  (let ((value (event-value event%)))
		     (trace-item "listener: " proc " for event: " event-object.name " with value: " value)
		     (proc value)))
	       event%.listeners)))))

;;; Build the extended environment, and update machine's environment
(define (%machine-append-program! machine::hiphop-machine% program::hiphop-program)
   (let ((program-copy::hiphop-program (hiphop-copy program))
	 (dynpar (hiphop-get-element-by-id machine.program.reactive-code '__dynpar)))
      (unless dynpar
	 (raise (instantiate::&hiphop-internal-error
		   (proc "%machine-append-program")
		   (msg "Cannot find element with id '__dynpar")
		   (obj #f))))
      (hiphop-append-child! dynpar
	 (instantiate::hiphop-sequence
	    (source-location program-copy.source-location)
	    (branches
	       (vector
		  ;; We have to set-up a boot pause, initially activated
		  (instantiate::hiphop-pause
		     (source-location program-copy.source-location)
		     (id (gensym 'dynpause))
		     (selected? #t))
		  ;; The actual user's program, putted into a run&
		  (instantiate::hiphop-run
		     (source-location program-copy.source-location)
		     (name program-copy.name)
		     (bindings '())
		     (program program-copy))))))
      (with-access::hiphop-parallel-dynamic dynpar (%env-builder)
	 (%env-builder))
      (machine-initialize! machine)
      #unspecified))
