(module __hiphop_must&can

   (import
      __hiphop_ast
      __hiphop_debug
      __hiphop_dom
      __hiphop_types
      __hiphop_environment
      __hiphop_events
      __hiphop_host_expressions
      __hiphop_utils)

   (export  (React! program::hiphop-program env::hiphop-environment%)))

;; React: Compute Must then Cannot as many times as needed until we
;; found a fixpoint or exceed the max loop iteration guard (count)
(define (React! program::hiphop-program env::hiphop-environment%)
   (let loop ((count 0))
      (set! env.%altered? #f)
      (let (; Compute Must
	    (must (Must program env))
	    ; We always compute Cannot in order to set the 'absent' status for not emitted events
	    (cannot (Cannot program env count)))
	 ;; If Must returns 0 or 1, the reaction is terminated
	 (when (not (or (Terminated? must) (Paused? must)))
	    ; Otherwise, we check if we should start another Must&Cannot computation
	    (cond
	       ((>fx count 9999) ; Sanity check to avoid stack depletion
		(raise (instantiate::&hiphop-internal-error
			  (proc "hiphop-react!")
			  (msg "Too much recursion")
			  (obj must))))
	       ;; Either must or cannot has made some progress, so loop
	       (env.%altered?
		(loop (+fx count 1)))
	       ;; If not must nor cannot made progress, there is a causality error.
	       (else
		(raise (instantiate::&hiphop-causality-error
			  (proc "hiphop-react!")
			  (msg "Cannot react. Unresolved program")
			  (obj must)))))))))

;;; Must: hiphop-program -> hiphop-environment% -> ?
;; Simple wrapper for MustResume on the program's reactive-code
(define (Must program::hiphop-program env::hiphop-environment%)
   ; Must always resume an execution
   ; (that's why the boot pause is selected for the first execution)
   (MustResume program.reactive-code env))


;;; Cannot: hiphop-program -> hiphop-environment% -> bool
;; Calls 'Can' on program's reactive-code, compute the complementary set
;; and mark event as absent if needed.
;; Returns true of some changes were made to the environment
(define (Cannot program::hiphop-program env::hiphop-environment% mark::bint)
   (with-trace 5 'Cannot
      (Can program.reactive-code env mark)
      (hashtable-for-each env.%emit-table
	 (lambda (key::pair emit::hiphop-emit)
	    ;; For each reachable emit statement that has not yet
	    ;; been executed (by Must) and has not been seen by the
	    ;; last Can iteration, then mark it as unreachable
	    (when (and emit.reachable? (not emit.flagged?) (<fx emit.%last-seen mark))
	       (trace-item "Mark #" mark "; Emit '" emit.id "' " key " cannot be reached (last-seen: " emit.%last-seen "). " emit)
	       (when (emit-unmark! emit)
		  (set! env.%altered? #t)))))))

;;; Potential functions: Must, Can
;; Following Compiling Esterel, chapter 4
;; Side-effect free?
(define-generic (MustResume::bint s::hiphop-ast% env::hiphop-environment%))
(define-generic (MustGo::bint     s::hiphop-ast% env::hiphop-environment%))
(define-generic (Can::bint        s::hiphop-ast% env::hiphop-environment% mark::bint . extras))

;;; Nothing
(define-method (MustGo s::hiphop-nothing env)
   (with-trace *hiphop-debug-must-go* 'Must/nothing&
      (Return 0)))
(define-method (Can  s::hiphop-nothing env mark . extras)
   (with-trace *hiphop-debug-can* 'Can/nothing&
      (Return 0)))

;;; Emit
(define-method (MustGo s::hiphop-emit env::hiphop-environment%)
   (with-trace *hiphop-debug-must-go* 'Must/emit&
      (with-access::hiphop-emit s (event (emit-index index) flagged? valued? expr)
	 (bind-exit (return)
	    (when flagged?
	       (raise (instantiate::&hiphop-internal-error
			 (proc "Must/emit")
			 (msg  "Already executed statement")
			 (obj s))))
	    (with-access::hiphop-event event (name event%)
	       (with-access::hiphop-event% event%
		     ((event-index index) (event-status status) emitted emitted-count)
		  (trace-item "Emitting #" emit-index ": event #" event-index ": " name)
		  (when (member emit-index emitted)
		     (raise (instantiate::&hiphop-internal-error
			       (proc "Must/emit")
			       (msg "Already executed emit statement")
			       (obj s))))
		  (if valued?
		      (multiple-value-bind (resolved? value)
			 (hiphop-host-expression-compute! s.expr env)
			 (if resolved?
			     (event-emit!/value event% value)
			     (return 0)))
		      (event-emit! event%))
		  (set! flagged? #t)
		  (set! emitted-count (+fx emitted-count 1))
		  (trace-item "self: " s.id)
		  (set! env.%altered? #t)
		  (Return 0)))))))
(define-method (MustResume s::hiphop-emit env)
   (when s.flagged?
      (raise (instantiate::&hiphop-internal-error
		(proc "Must/Resume/emit")
		(msg "Attempting to resume a terminated emit statement!")
		(obj s))))
   (MustGo s env))
(define-method (Can emit::hiphop-emit env mark . extras)
   (with-trace *hiphop-debug-can* 'Can/emit&
      (trace-fields emit id index reachable?)
      (with-access::hiphop-event emit.event (name event%)
	 (with-access::hiphop-event% event% ((event-index index))
	    (trace-item "Can emit event #" event-index ": " name ". Mark #" mark)
	    (set! emit.%last-seen mark)
	    (Return 0)))))

;;; Pause
(define-method (MustGo pause::hiphop-pause env::hiphop-environment%)
   (with-trace *hiphop-debug-must-go* 'Must/Go/pause&
      (trace-item "pause.id: " pause.id "; pausing...")
      (pause-table-set! "Must/Go/pause&" env pause.id #t)
      (Return 1)))
(define-method (MustResume pause::hiphop-pause env::hiphop-environment%)
   (with-trace *hiphop-debug-must-resume* 'Must/Resume/pause&
      (trace-item "pause.id: " pause.id "; resuming...")
      (pause-table-set! "Must/Resume/pause&" env pause.id #f)
      (Return 0)))
(define-method (Can pause::hiphop-pause env mark . extras)
   (with-trace *hiphop-debug-can* 'Can/pause
      (trace-item "pause.id: " pause.id "; " (pause-table-get "Can/pause&" env pause.id) "; " pause.selected?)
      (Return (if (and pause.selected? (not (ignore-decorations? extras))) 0 1))))

;;; Goto Pause
; Behave like pause for Go. Error if resume
(define-method (MustGo pause::hiphop-goto-pause env::hiphop-environment%)
   (with-trace *hiphop-debug-must-go* 'Must/Go/gogo-pause&
      (trace-item "pause.id: " pause.id "; pausing...")
      (pause-table-set! "Must/Go/goto-pause&" env pause.id #t)
      (Return 1)))
(define-method (MustResume pause::hiphop-goto-pause env::hiphop-environment%)
   (raise (instantiate::&hiphop-internal-error
	     (proc "Must/Resume/goto-pause&")
	     (msg "Attempting to resume a goto-pause statement")
	     (obj pause)))
   (Return 1))
(define-method (Can pause::hiphop-goto-pause env mark . extras)
   (with-trace *hiphop-debug-can* 'Can/goto-pause
      (trace-item "pause.id: " pause.id)
      (Return 1)))

;;; If
(define-method (MustGo if%::hiphop-if env)
   (with-trace *hiphop-debug-must-go* 'Must/Go/if
      (trace-fields if% id branch)
      (if (eq? if%.branch 'none)
	  (let ((resolved? (hiphop-host-expression-resolve! if%.test env)))
	     (if resolved?
		 (let ((test-value (hiphop-host-expression-eval! if%.test env)))
		    (set! if%.branch (if test-value 'then 'else))
		    (case if%.branch
		       ((then)
			(MustGo if%.then env))
		       ((else)
			(MustGo if%.else env))))
		 (begin
		    (set! if%.branch 'rego)
		    0)))
	  (raise (instantiate::&hiphop-internal-error
		    (proc "Must/Go/if")
		    (msg "Already selected if")
		    (obj if%))))))
(define-method (MustResume s::hiphop-if env)
   (with-trace *hiphop-debug-must-resume* 'Must/Resume/if
      (trace-fields s id branch)
      (with-access::hiphop-if s (branch (if-then then) (if-else else))
	 (case branch
	    ((rego)
	     (set! s.branch 'none)
	     (MustGo s env))
	    ((then)
	     (trace-item "Resuming then branch")
	     (let ((return (MustResume if-then env)))
		(trace-item "then-ret: " return)
		return))
	    ((else)
	     (trace-item "Resuming else branch")
	     (let ((return (MustResume if-else env)))
		(trace-item "then-ret: " return)
		return))
	    (else
	     (raise (instantiate::&hiphop-internal-error
		       (proc "Must/Resume/if")
		       (msg "Illegal selected branch")
		       (obj branch))))))))
(define-method (Can  s::hiphop-if env mark . extras)
   (with-trace *hiphop-debug-can* 'Can/if
      (with-access::hiphop-if s ((if-then then) (if-else else))
	 (let ((ignore-states? (ignore-decorations? extras)))
	    (trace-item "branch: " s.branch)
	    (if ignore-states?
		(Union (Can if-then env mark ignore-states?) (Can if-else env mark ignore-states?))
		(let ((branch s.branch))
		   (case branch
		      ((none rego) (Union (Can if-then env mark) (Can if-else env mark)))
		      ((then) (Can if-then env mark))
		      ((else) (Can if-else env mark))
		      (else   (raise (instantiate::&hiphop-internal-error
					(proc "Can/if&")
					(msg "Illegal selected branch")
					(obj s)))))))))))

;;; Suspend
(define-method (MustGo s::hiphop-suspend env)
   (with-trace *hiphop-debug-must-go* 'Must/Go/suspend
      (trace-fields s id %rego?)
      (set! s.%rego? #f)
      (let ((body (MustGo s.body env)))
	 (trace-item "body-ret: " body)
	 (when (zerofx? body)
	    (set! s.%rego? #t))
	 body)))
(define-method (MustResume s::hiphop-suspend env)
   (with-trace *hiphop-debug-must-resume* 'Must/Resume/suspend
      (trace-fields s id)
      (if s.%rego?
	  (let ((res (MustResume s.body env)))
	     (unless (zerofx? res)
		(set! s.%rego? #f))
	     res)
	  (multiple-value-bind (resolved? value)
	     (hiphop-host-expression-compute! s.event-test env)
	     (if resolved?
		 (if value ; then acts as pause
		     (begin
			(trace-item "Suspended")
			(Return 1))
		     (begin ; Propagate to body
			(trace-item "Resuming body")
			(MustResume s.body env)))
		 0)))))
(define-method (Can  s::hiphop-suspend env mark . extras)
   (with-trace *hiphop-debug-can* 'Can/suspend&
      (if (hiphop-host-expression-computed? s.event-test)
	  (multiple-value-bind (resolved? value)
	     (hiphop-host-expression-compute! s.event-test env)
	     (hiphop-host-expression-reset! s.event-test)
	     (if resolved?
		 (if value
		     (Return 1)
		     (Can s.body env mark (ignore-decorations? extras)))
		 (Union (Return 1)
		    (Can s.body env mark (ignore-decorations? extras)))))
	  (Union (Return 1)
	     (Can s.body env mark (ignore-decorations? extras))))))

;;; Sequence
; Recursively propagates Go to the next branch
(define-method (MustGo seq::hiphop-sequence env::hiphop-environment%)
   (with-trace *hiphop-debug-must-go* 'Must/Go/seq&
      (trace-fields seq id current-branch)
      (trace-item "seq.branches.length: " (vector-length seq.branches))
      (let ((current-branch seq.current-branch)
	    (last-branch (-fx (vector-length seq.branches) 1)))
	 ; First check if we are terminated
	 (if (=fx current-branch last-branch)
	     (Return 0)
	     (begin
		; If not, starts the next branch, and if it terminates,
		; recursively calls Go on self, which will select the next branch
		(set! seq.current-branch (+fx current-branch 1))
		(let* ((branch (vector-ref seq.branches seq.current-branch))
		       (result  (MustGo branch env)))
		   (trace-fields seq current-branch)
		   (trace-item "branch-result: " result)
		   (if (Terminated? result)
		       (begin
			  (Reset! branch env)
			  (Union/right result (MustGo seq env)))
		       result)))))))
(define-method (MustResume s::hiphop-sequence env)
   (with-trace *hiphop-debug-must-resume* 'Must/Resume/seq&
      (trace-item "id: " s.id "; current-branch: " s.current-branch)
      (let ((current s.current-branch))
	 (if (=fx current -1)
	     (raise (instantiate::&hiphop-internal-error
		       (proc "Resume/seq&")
		       (msg "Attempting to resume a non-stopped statement")
		       (obj s)))
	     (let* ((branch (vector-ref s.branches current))
		    (result (MustResume branch env)))
		(if (Terminated? result)
		    (Union/right result (MustGo s env))
		    result))))))
(define-method (Can s::hiphop-sequence env mark . extras)
   (with-trace *hiphop-debug-can* 'Can/seq&
      (let ((l (vector-length s.branches))
	    (out #f)
	    (ignore-states? (ignore-decorations? extras)))
	 (let loop ((c (if (or ignore-states? (<fx s.current-branch 0))
			   0 s.current-branch)))
	    (if (<fx c l)
		(let ((result (Can (vector-ref s.branches c) env mark ignore-states?)))
		   (set! out (if out (Union/!zero out result) result))
		   (trace-item "result: " result)
		   (if (Terminated? result)
		       (loop (+fx c 1))
		       out))
		(or out (Return 0)))))))

;;; Loop
(define-method (MustGo loop::hiphop-loop env::hiphop-environment%)
   (with-trace *hiphop-debug-must-go* 'Must/Go/loop&
      (trace-fields loop id)
      ; Recursively calls Go on the loop's body
      (let ((go (MustGo loop.body env)))
	 (trace-item "body-ret: " go)
	 ; If the body terminates, raise an exception
	 (when (Terminated? go)
	    (raise (instantiate::&hiphop-causality-error
		      (proc "loop&")
		      (msg "Instantaneous loop")
		      (obj loop)
		      (location loop.source-location))))
	 go)))
(define-method (MustResume loop::hiphop-loop env::hiphop-environment%)
   (with-trace *hiphop-debug-must-resume* 'Must/Resume/loop&
      (trace-fields loop id)
      ; Recursively resume the loop's body
      (let ((resume (MustResume loop.body env)))
	 ; Check if the body terminates
	 (if (Terminated? resume)
	     (begin
		(trace-item "Restarting loop, body-ret: " resume)
		; The resume has terminated, so clear all decorations (Reset!)
		(Reset! loop.body env)
		; Then calls Go on self (Go does the instantaneous loop check)
		(Union/right resume (MustGo loop env)))
	     resume))))
(define-method (Can loop::hiphop-loop env mark . extras)
   (with-trace *hiphop-debug-can* 'Can/loop&
      (trace-fields loop id)
      (let ((ignore-states? (ignore-decorations? extras)))
	 (let ((res (Can loop.body env mark ignore-states?)))
	    (trace-item "result: " res)
	    (if (Terminated? res)
		(Can loop.body env mark :ignore-decorations)
		res)))))

;;; Parallel
; Go over all branches
(define-method (MustGo par::hiphop-parallel env::hiphop-environment%)
   (with-trace *hiphop-debug-must-go* 'Must/Go/par&
      (trace-item "id: " par.id)
      (let ((out #f)
	    (l (vector-length par.branches)))
	 (let loop ((i 0))
	    (if (<fx i l)
		(let ((res-i (MustGo (vector-ref par.branches i) env)))
		   (trace-item "branch " i "; ret=" res-i)
		   (when (>fx res-i 0)
		      ; Stop the branch if the branch has returned
		      (vector-set! par.active-branches i #f))
		   (if out
		       (set! out (Union/max out res-i))
		       (set! out res-i))
		   (loop (+fx i 1)))
		(or out (Return 0)))))))
(define-method (MustResume s::hiphop-parallel env)
   (with-trace *hiphop-debug-must-resume* 'Must/Resume/par&
      (trace-item "id: " s.id)
      (let ((out #f)
	    (l (vector-length s.branches)))
	 ; Iteration on each branch for the parallel statement
	 (let loop ((i 0))
	    (if (<fx i l)
		(begin
		   (trace-item "Branch " i ", active? " (vector-ref s.active-branches i))
		   ; Skip the i-branch if it has been marked as inactive
		   (when (vector-ref s.active-branches i)
		      (let* ((branch (vector-ref s.branches i))
			     (res (MustResume branch env)))
			 (trace-item "ret=" res)
			 ; If the branch has terminated, then we mark it has inactive
			 (when (>fx res 0)
			    ; Stop the branch if the branch has returned
			    (trace-item "branch has terminated")
			    (vector-set! s.active-branches i #f))
			 (trace-item "res: " res)
			 ; Compute the Union of the previous branches and the current one
			 (if out
			     (set! out (Union/max out res))
			     (set! out res))))
		   (loop (+fx i 1)))
		(begin
		   (trace-item "out: " out)
		   ; If all branches have already terminated, act as nothing&
		   (or out (Return 0))))))))
(define-method (Can s::hiphop-parallel env mark . extras)
   (with-trace *hiphop-debug-can* 'Can/par&
      (trace-item "par.id: " s.id)
      (let ((out #f)
	    (l (vector-length s.branches))
	    (ignore-states? (ignore-decorations? extras)))
	 (let loop ((i 0))
	    (if (<fx i l)
		(begin
		   (when (vector-ref s.active-branches i)
		      (let ((res (Can (vector-ref s.branches i) env mark ignore-states?)))
			 (if out
			     (set! out (Union/max out res))
			     (set! out res))))
		   (loop (+fx i 1)))
		(or out 0))))))

(define-method (MustGo genpar::hiphop-genpar env::hiphop-environment%)
   (with-trace *hiphop-debug-must-go* 'Must/Go/genpar&
      (trace-fields genpar id)
      (when genpar.expanded?
	 (raise (instantiate::&hiphop-internal-error
		   (proc "MustGo/genpar&")
		   (msg "Received Go on an already started expression")
		   (obj genpar))))
      ;; First, resolve all host expressions
      (let ((host-resolved (map (lambda (host-expr)
				   (hiphop-host-expression-resolve! host-expr env))
			      genpar.host-expressions)))
	 (trace-item "host-resolved: " host-resolved)
	 (if (every? (lambda (bool) bool) host-resolved)
	     ;; If all expressions are resolved, compute their values
	     (let ((host-values (map (lambda (host-expr)
					(hiphop-host-expression-eval! host-expr env))
				   genpar.host-expressions)))
		;; Build all new branches using the values
		(let ((branches (apply cartesian-map genpar.generator host-values)))
		   (trace-item "I just have generated " (length branches) " new branches")
		   ;; Fill the par& fields for genpar&
		   (set! genpar.branches (list->vector branches))
		   (set! genpar.active-branches (make-vector (vector-length genpar.branches) #t))
		   ;; Build the environment for each new branches
		   (my-vector-for-each genpar.%env-builder genpar.branches)
		   ;; Update the status as expanded
		   (set! genpar.expanded? #t)
		   ;; Initialize new programs if any
		   (machine-initialize! env.current-machine)
		   ;; Act as par& from now on
		   (call-next-method)))
	     ;; Some events values are still missing, will come back here a bit later
	     0))))
(define-method (MustResume genpar::hiphop-genpar env::hiphop-environment%)
   (with-trace *hiphop-debug-must-resume* 'Must/Resume/genpar&
      (trace-fields genpar id expanded?)
      (if genpar.expanded?
	  (call-next-method)
	  (MustGo genpar env))))
(define-method (Can genpar::hiphop-genpar env::hiphop-environment% mark . extras)
   (with-trace *hiphop-debug-can* 'Can/genpar&
      (unless genpar.can-nil-instance
	 (raise (instantiate::&hiphop-internal-error
		   (proc "Can/genpar")
		   (msg "The can-nil-instance field has not been initialized")
		   (obj genpar))))
      (Can genpar.can-nil-instance env mark (ignore-decorations? extras))))

;;; Algorithm
;; On Go, acts as pause
;; On Resume, compute new branches, if any (can be a new instance of genpar&)
;;  Then call Go on this new branches
;; Call Resume on pre-existing branches
;; Append pre-existing branches and new branches
;; Return (max branches 1)
(define-method (MustGo dgp::hiphop-dyngenpar env::hiphop-environment%)
   (with-trace *hiphop-debug-must-go* 'Must/Go/dyngenpar&
      (set! dgp.has-started? #t)
      (set! dgp.expanded? #t)
      (Return 1)))
(define-method (MustResume dgp::hiphop-dyngenpar env::hiphop-environment%)
   (with-trace *hiphop-debug-must-resume* 'Must/Resume/dyngenpar&
      (if (hiphop-host-expression-resolve! dgp.when env)
	  (begin
	     (unless dgp.expanded?
		;; Compute when expression
		(when (hiphop-host-expression-eval! dgp.when env)
		   ;; Then duplicate genpar
		   (let ((new-genpar (hiphop-copy-ast dgp.genpar)))
		      ;; Add it to self
		      (hiphop-append-child! dgp new-genpar)
		      ;; Mark the newly created branch as active
		      (vector-set! dgp.active-branches (-fx (vector-length dgp.branches) 1) #t))
		   (set! dgp.has-new-branch? #t)
		   ;; Rebuild the environment
		   (dgp.%env-builder)
		   ;; Initialize new programs if any
		   (machine-initialize! env.current-machine))
		;; Mark the dyngenpar as expanded for the current reaction
		(set! dgp.expanded? #t))
	     ;; Call next method if selected branches
	     (if dgp.has-new-branch?
		 (Union/max
		    (if dgp.selected-branches?
			(call-next-method)
			(MustGo (vector-ref dgp.branches (-fx (vector-length dgp.branches) 1)) env))
		    (Return 1))
		 (Return 1)))
	  0)))
(define-method (Can dgp::hiphop-dyngenpar env::hiphop-environment% mark . extras)
   (with-trace *hiphop-debug-must-resume* 'Can/dyngenpar&
      (let ((ignore-states? (ignore-decorations? extras)))
	 (if ignore-states?
	     (Can dgp.genpar env mark ignore-states?)
	     (if dgp.has-started?
		 (if dgp.expanded?
		     (call-next-method)
		     (Union
			(call-next-method)
			(Can dgp.genpar env mark)))
		 (Return 1))))))

;;; Trap
;; Go/Resume on trap's body, then analyse k
(define-method (MustGo trap::hiphop-trap env::hiphop-environment%)
   (with-trace *hiphop-debug-must-go* 'Must/Go/trap&
      (trace-fields trap id name)
      (let ((go (MustGo trap.body env)))
	 (trace-item "body-ret " go)
	 (when (or (Terminated? go)
		   (Exited? go))
	    (trace-item "My body has terminated or exited, resetting...")
	    (Reset! trap.body env))
	 (Downshift go))))
(define-method (MustResume trap::hiphop-trap env::hiphop-environment%)
   (with-trace *hiphop-debug-must-resume* 'Must/Resume/trap&
      (trace-fields trap id name)
      (let ((resume (MustResume trap.body env)))
	 (trace-item "body-ret " resume)
	 (when (or (Terminated? resume)
		   (Exited? resume))
	    (Reset! trap.body env))
	 (Downshift resume))))
(define-method (Can trap::hiphop-trap env mark . extras)
   (with-trace *hiphop-debug-can* 'Can/trap&
      (trace-fields trap id name)
      (let ((ignore-states? (ignore-decorations? extras)))
	 (Downshift (Can trap.body env mark ignore-states?)))))

;;; Exit
(define-method (MustGo exit::hiphop-exit env)
   (with-trace *hiphop-debug-must-go* 'Must/exit&
      (trace-fields exit index trap-name)
      (Return exit.index)))
(define-method (Can exit::hiphop-exit env mark . extras)
   (with-trace *hiphop-debug-can* 'Can/exit&
      (trace-fields exit index trap-name)
      (Return exit.index)))

;;; Local
;; Go/Resume on local's body
(define-method (MustGo local::hiphop-local env::hiphop-environment%)
   ;(with-trace *hiphop-debug-must-go* 'Must/Go/local&
      (MustGo local.body env));)
(define-method (MustResume local::hiphop-local env::hiphop-environment%)
   ;(with-trace *hiphop-debug-must-resume* 'Must/Resume/local&
      (MustResume local.body env));)
(define-method (Can s::hiphop-local env mark . extras)
   ;(with-trace *hiphop-debug-can* 'Can/local&
      (let ((ignore-states? (ignore-decorations? extras)))
	 (Can s.body env mark ignore-states?)));)

;;; Atom
(define-method (MustGo atom::hiphop-atom env)
   (with-trace *hiphop-debug-must-go* 'Must/Go/atom&
      (trace-fields atom id ran? source-location)
      (unless atom.ran?
	 (trace-item "Running expression: " atom.expr)
	 (multiple-value-bind (resolved? value)
	    (hiphop-host-expression-compute! atom.expr env)
	    (if resolved?
		(begin
		   (set! atom.ran? #t)
		   (Return 0))
		0)))))
(define-method (MustResume atom::hiphop-atom env)
   (with-trace *hiphop-debug-must-resume* 'Must/Resume/atom&
      (trace-fields atom id ran? source-location)
      (if atom.ran?
	  (raise (instantiate::&hiphop-internal-error
		    (proc "Must/Resume/atom&")
		    (msg "Attempting to resume a terminated atom& expression")
		    (obj atom)))
	  (MustGo atom env))))
(define-method (Can  s::hiphop-atom env mark . extras)
   (with-trace *hiphop-debug-can* 'Can/atom&
      (Return 0)))


;;; Run
(define-method (MustGo s::hiphop-run env)
   (with-trace *hiphop-debug-must-go* 'Must/Go/run&
      (trace-item "program name: " s.program.name)
      (when s.name
	 (trace-item "renamed to: " s.name))
      (MustGo s.program.reactive-code env)))
(define-method (MustResume s::hiphop-run env)
   (with-trace *hiphop-debug-must-resume* 'Must/Resume/run&
      (trace-item "program name: " s.program.name)
      (when s.name
	 (trace-item "renamed to: " s.name))
      (MustResume s.program.reactive-code env)))
(define-method (Can  s::hiphop-run env mark . extras)
   (with-trace *hiphop-debug-can* 'Can/run&
      (trace-item "program name: " s.program.name)
      (when s.name
	 (trace-item "renamed to: " s.name))
      (let ((ignore-states? (ignore-decorations? extras)))
	 (Can s.program.reactive-code env mark ignore-states?))))


;;; private functions
(define-inline (Return code::bint)
   (bint->bv code))
(define-inline (Terminated?::bool return::bint)
   (bv-has-zero? return))
(define-inline (Paused?::bool return::bint)
   (not (zerofx? (bit-and (bint->bv 1) return))))
(define-inline (Exited?::bool return::bint . args)
   (let ((code (if (pair? args) (car args) 2)))
      (not (zerofx? (bit-and (bint->bv code) return)))))

(define (Union return-code-1::bint return-code-2::bint)
   (bv-union return-code-1 return-code-2))
(define (Union/!zero return-code-1::bint return-code-2::bint)
   (bv-union/!zero return-code-1 return-code-2))
(define (Union/right return-code-1::bint return-code-2::bint)
   return-code-2)
(define (Union/max return-code-1::bint return-code-2::bint)
   (bv-max return-code-1 return-code-2))

(define-inline (Downshift return-code::bint)
   (bv-downshift return-code))

(define-inline (ignore-decorations? args) ; arg might be a list
   (let loop ((l args))
      (when (pair? l)
	 (let ((f (car l)))
	    (if (eq? f :ignore-decorations)
		f
		(loop (cdr l)))))))

;;; Reset! do the same as clear-decorations! and "kills" selected pauses
(define-generic (Reset! s::hiphop-ast% env::hiphop-environment%))
(define-method (Reset! s::hiphop-nothing env) #f)
(define-method (Reset! s::hiphop-emit env::hiphop-environment%) #f)
(define-method (Reset! pause::hiphop-pause env::hiphop-environment%)
   (pause-table-set! "Reset!/pause" env pause.id #f)
   (set! pause.selected? #f))
(define-method (Reset! s::hiphop-goto-pause env) #f)
(define-method (Reset! if%::hiphop-if env::hiphop-environment%)
   '(case if%.branch
      ((then) (Reset! if%.then env))
      ((else) (Reset! if%.else env)))
   (Reset! if%.then env)
   (Reset! if%.else env)
   (set! if%.branch 'none))
(define-method (Reset! s::hiphop-suspend env)
   (Reset! s.body env))
(define-method (Reset! s::hiphop-sequence env::hiphop-environment%)
   (with-trace *hiphop-debug-reset* 'Reset!/seq
      (trace-fields s id)
      (set! s.current-branch -1)
      (let ((l (vector-length s.branches)))
	 (let loop ((i 0))
	    (when (<fx i l)
	       (Reset! (vector-ref s.branches i) env)
	       (loop (+fx i 1)))))))
(define-method (Reset! s::hiphop-loop env::hiphop-environment%)
   (with-access::hiphop-loop s (body)
      (Reset! body env)))
(define-method (Reset! s::hiphop-parallel env::hiphop-environment%)
   (let ((l (vector-length s.branches)))
      (let loop ((i 0))
	 (when (<fx i l)
	    (vector-set! s.active-branches i #t)
	    (Reset! (vector-ref s.branches i) env)
	    (loop (+fx i 1))))))
(define-method (Reset! s::hiphop-genpar env::hiphop-environment%)
   (set! s.expanded? #f)
   ;(set! s.branches (vector))
   ;(set! s.active-branches (vector))
   (Reset! s.can-nil-instance env))
(define-method (Reset! dgp::hiphop-dyngenpar env::hiphop-environment%)
   (clear-decorations! dgp)
   (set! dgp.has-started? #f)
   (set! dgp.branches (vector))
   (set! dgp.active-branches (vector))
   (Reset! dgp.genpar env))
(define-method (Reset! s::hiphop-trap env::hiphop-environment%)
   (with-access::hiphop-trap s (body)
      (Reset! body env)))
(define-method (Reset! s::hiphop-exit env) #f)
(define-method (Reset! s::hiphop-local env::hiphop-environment%)
   (with-access::hiphop-local s (body)
      (Reset! body env)))
(define-method (Reset! s::hiphop-atom env::hiphop-environment%)
   (with-access::hiphop-atom s (ran?)
      (set! ran? #f)))
(define-method (Reset! run::hiphop-run env::hiphop-environment%)
   (Reset! run.program.reactive-code env))

(define (emit-unmark!::bool s::hiphop-emit)
   (with-trace 21 'Prune/emit
      (when s.flagged?
	 (raise (instantiate::&hiphop-internal-error
		   (proc "Prune")
		   (msg "Attempting to prune a previously executed statement")
		   (obj s))))
      (let ((event::hiphop-event s.event))
	 (when (and s.reachable? (not s.flagged?))
	    (trace-item "pruning emitter " s.index " for event #" event.event%.index)
	    (trace-item "emit: " s)
	    (trace-item "event%: " event.event%)
	    (set! s.reachable? #f)
	    (set! event.event%.reachable-count (-fx event.event%.reachable-count 1))
	    (event-update-status! event.event%)
	    #t))))
