(module __hiphop_utils
   (import __hiphop_types)

   ;; Macros
   (export (macro trace-fields))

   ;; Error/Warning utilities
   (export
      (hiphop-warn loc proc . args)
      (hiphop-type-error proc expected obj #!optional loc other)
      (hiphop-compile-error loc proc msg obj))

   ;; Utilities
   (export
      (my-vector-for-each ::procedure ::vector)
      (assev event-name::bstring event-list::pair-nil))

   ;; bitvector operations
   (export (bint->bv::bint ::bint)
           (bv-has-zero?::bool ::bint)
	   (bv-union::bint ::bint ::bint)
	   (bv-union/!zero::bint ::bint ::bint)
	   (bv-inter::bint ::bint ::bint)
	   (bv-max::bint ::bint ::bint)
	   (bv-downshift::bint ::bint))

   ;; "Cartesian"-map
   (export (cartesian-map::pair-nil ::procedure . rest)))

;; For tracing fields of bigloo objects
(define-macro (trace-fields object . fields)
   `(begin
       ,@(map (lambda (field)
		 `(trace-item ,(format "~a.~a: " object field) (-> ,object ,field)))
	    fields)))

;; Helper for calling warn or warn/loc
(define (hiphop-warn loc proc . args)
   (let ((arg (apply string-append args)))
      (cond-expand
	 (scheme2js
	    (when loc
	       (tprint "*** WARNING:" (format "~a" loc) " " proc ":\n" arg)))
	 (else
	  (if loc
	      (warning/location (cadr loc) (caddr loc) proc arg)
	      (warning proc arg))))))

;; Type error management
(define (hiphop-type-error proc expected obj #!optional loc other)
   (let ((msg (string-append "Type `" expected "' expected, `" (typeof obj) "' provided")))
      (hiphop-compile-error loc proc msg (if other (list obj "Other arguments:" other) obj))))

(define (hiphop-compile-error loc proc msg obj)
   (match-case loc
      ((at ?fname ?point)
       (instantiate::&hiphop-compile-error
	  (fname fname)
	  (location point)
	  (proc proc)
	  (msg msg)
	  (obj obj)))
      (else
       (instantiate::&hiphop-compile-error
	  (proc proc)
	  (msg msg)
	  (obj obj)))))

;;; for-each procedure that operate on vectors
(define (my-vector-for-each proc vect)
   (let ((len (vector-length vect)))
      (let loop ((index 0))
	 (when (< index len)
	    (proc (vector-ref vect index))
	    (loop (+fx index 1))))))

;;; Search through a list of event the event named event-name
;; assev is not a good name, a previous version was very similar to assoc
(define (assev event-name::bstring event-list::pair-nil)
   (let loop ((l event-list))
      (when (pair? l)
	 (let ((event (car l)))
	    (if (isa? event hiphop-event)
		(with-access::hiphop-event event (name)
		   (if (string=? name event-name)
		       event
		       (loop (cdr l))))
		(raise (instantiate::&hiphop-internal-error
			  (proc "assev")
			  (msg "Not an hiphop-event object")
			  (obj event))))))))


;;; Functions to manages integer sets as bit vectors

;; Convert a fixnum n into a bit vector, whose nth bit it set to 1
(define (bint->bv::bint n::bint)
   ;; 32 bits implementation of fixnum reserve 2 bits for tagging.
   (if (>= n 30)
       (raise (instantiate::&hiphop-internal-error
		 (proc "int->bv")
		 (msg "Out of range value for bitvectors")
		 (obj n)))
       (bit-lsh 1 n))); 1 << n

;; Test if the 0th bit is 1 (the ba contains 0)
(define (bv-has-zero?::bool A::bint)
   (not (zerofx? (bit-and 1 A)))) ; ! ( (A & 1) == 0 )

;; Union of two bit vectors
(define (bv-union::bint A::bint B::bint)
   (bit-or A B))

;; Union of two bit vectors, but removing zero from the first one
(define (bv-union/!zero::bint A::bint B::bint)
   (bit-or
      (bit-and -2 A) ; set the 0-th bit to 0
      B))

;; Intersection of two bit vectors
(define (bv-inter::bint A::bint B::bint)
   (bit-and A B))

;; Max of two bit vectors
(define (bv-max::bint A::bint B::bint)
   (cond
      ((zerofx? A) A)
      ((zerofx? B) B)
      (else
       (bit-and
	  (bit-or A B) ; A|B
	  (bit-and
	     (bit-or A (- A))      ; A|-A
	     (bit-or B (- B))))))) ; B|-B

;; Downshifting a bit vector
(define (bv-downshift::bint A::bint)
   (let ((zero-mask  5)  ; 0..00101, check if the 0th or 2nd bit is set
	 (pause-mask 2)  ; 0..00010, check if the 1st bit is set
	 (rest-mask -8)) ; 1..11000, the rest of the bit vector
      (let ((result (bit-ursh (bit-and A rest-mask) 1))) ; R = (A & -8) >> 1
	 (unless (zerofx? (bit-and A zero-mask))
	    ; if ( A & 101 ) then R |= 1 fi
	    (set! result (bit-or result 1)))
	 (unless (zerofx? (bit-and A pause-mask))
	    ; if ( A & 10 ) then R |= 2 fi
	    (set! result (bit-or result 2)))
	 result)))


;*---------------------------------------------------------------------*/
;*    Cartesian-map implementation ...                                 */
;*---------------------------------------------------------------------*/

;; Recursive flattening of list, NOT properly tail-rec
(define (flatten l)
   (let ((flattened '()))
      (let loop ((to-flatten l))
	 (when (pair? to-flatten)
	    (if (pair? (car to-flatten))
		(loop (append (car to-flatten) (cdr to-flatten)))
		(begin
		   (set! flattened (cons (car to-flatten) flattened))
		   (loop (cdr to-flatten))))))
      (reverse! flattened)))

;; Cartesian product of 2 lists
(define (cartesian-2 l1 l2)
  (apply append
     (map
	(lambda (e1)
	   (map (lambda (e2) (list e1 e2)) l2))
	l1)))

;; Cartesian product of n list, done 2-by-2, from leftmost list
(define (cartesian-n l1 l2 . l)
   (let loop ((c-2 (cartesian-2 l1 l2))
	      (rest l))
      (if (pair? rest)
	  (loop (map flatten (cartesian-2 c-2 (car rest))) (cdr rest))
	  c-2)))

;; Exported
;; Cartesian-map
(define (cartesian-map proc . l)
   (if (pair? l)
       (let ((l1 (car l))
	     (r1 (cdr l)))
	  (if (pair? r1)
	      (let ((l2 (car r1))
		    (r2 (cdr r1)))
		 (map (lambda (v) (apply proc v)) (apply cartesian-n l1 l2 r2)))
	      (map proc l1)))
       '()))
